//////////////////////////////////////////////////////////////////////////////
//   Package:          CAPD
//   Subpackage:
/// @addtogroup intervals
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file intervalFriend.h
///
/// Interval Arithmetics - friend functions template declarations
///
/// @author Tomasz Kapela   @date 11-01-2006
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) Tomasz Kapela 2006
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _INTERVAL_FRIEND_H_
#define _INTERVAL_FRIEND_H_

#include <iostream>

namespace chomp{
namespace intervals{

template < typename T_Bound, typename T_Rnd = T_Bound >
class Interval;

//////////////////////////////////////////////////////////////////////////
//     access to interval ends
//////////////////////////////////////////////////////////////////////////

/// returns interval containing left end
template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd >  left( const Interval< T_Bound, T_Rnd >  & A_iv )
{
   return Interval<T_Bound, T_Rnd>( A_iv.leftBound());
}

/// returns interval containing right end
template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd >  right( const Interval< T_Bound, T_Rnd >  & A_iv )
{
   return Interval<T_Bound, T_Rnd>( A_iv.rightBound() );
}

/// returns left end
template < typename T_Bound, typename T_Rnd> 
inline T_Bound  leftBound( const Interval< T_Bound, T_Rnd >  & A_iv )
{
  return  A_iv.leftBound();
}

/// returns right end
template < typename T_Bound, typename T_Rnd> 
inline T_Bound  rightBound( const Interval< T_Bound, T_Rnd >  & A_iv )
{
  return  A_iv.rightBound();
}

//////////////////////////////////////////////////////////////////////////
//     RELATIONS
//////////////////////////////////////////////////////////////////////////

/// the relation  ==
template < typename T_Bound, typename T_Rnd> 
inline bool operator == (const Interval< T_Bound, T_Rnd > & A_iv1, 
                         const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return ((A_iv1.leftBound()==A_iv2.leftBound()) && (A_iv1.rightBound()==A_iv2.rightBound()));
}

/// weak inequality relation <=
template < typename T_Bound, typename T_Rnd> 
inline bool operator <= (const Interval< T_Bound, T_Rnd > & A_iv1, 
                         const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return (A_iv1.rightBound() <= A_iv2.leftBound());
}


/// weak inequality relation >=
template < typename T_Bound, typename T_Rnd> 
inline bool operator >= (const Interval< T_Bound, T_Rnd > & A_iv1, 
                         const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return (A_iv1.leftBound()>= A_iv2.rightBound() );
}

/// strong inequality <
template < typename T_Bound, typename T_Rnd> 
inline bool operator < (const Interval< T_Bound, T_Rnd > & A_iv1, 
                        const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return (A_iv1.rightBound() < A_iv2.leftBound());
}

/// strong inequality relation >
template < typename T_Bound, typename T_Rnd> 
inline bool operator > (const Interval< T_Bound, T_Rnd > & A_iv1, 
                        const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return (A_iv1.leftBound()> A_iv2.rightBound());
}

/// the relation  !=  ( as a negation of  ==)
template < typename T_Bound, typename T_Rnd> 
inline bool operator != (const Interval< T_Bound, T_Rnd > & A_iv1, 
                         const Interval< T_Bound, T_Rnd > & A_iv2)
{
   return ((A_iv1.leftBound()!= A_iv2.leftBound()) || (A_iv1.rightBound() != A_iv2.rightBound()));
}


//////////////////////////////////////////////////////////////////////////
//     INPUT and OUTPUT
//////////////////////////////////////////////////////////////////////////

/// sending an Interval to a stream
template < typename T_Bound, typename T_Rnd> 
std::ostream & operator << (std::ostream & s, 
                            const Interval< T_Bound, T_Rnd > & A_iv);

/// reading an Interval from a stream
template < typename T_Bound, typename T_Rnd> 
std::istream & operator >> (std::istream & s, 
                            Interval< T_Bound, T_Rnd > & A_iv);

//////////////////////////////////////////////////////////////////////////
//     ARITHMETIC OPERATORS
//////////////////////////////////////////////////////////////////////////

/// unary operator -
template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator-(const Interval< T_Bound, T_Rnd> & A_iv)
{
  return(Interval< T_Bound, T_Rnd>(-A_iv.rightBound(),-A_iv.leftBound()));
}// unary operator -


template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> operator + (const Interval< T_Bound, T_Rnd>& A_iv1, 
                                      const Interval< T_Bound, T_Rnd>& A_iv2);

template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> operator - (const Interval< T_Bound, T_Rnd>& A_iv1, 
                                      const Interval< T_Bound, T_Rnd>& A_iv2);

template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> operator * (const Interval< T_Bound, T_Rnd>& A_iv1, 
                                      const Interval< T_Bound, T_Rnd>& A_iv2);

template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> operator / (const Interval< T_Bound, T_Rnd>& A_iv1, 
                                      const Interval< T_Bound, T_Rnd>& A_iv2);

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator ^ (const Interval< T_Bound, T_Rnd>& A_iv1, 
                                             int i)
{
   return power(A_iv1, i);
}


//////////////////////////////////////////////////////////////////////////
//     ARITHMETIC OPERATORS  between Interval and double
//////////////////////////////////////////////////////////////////////////
/*
template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator + (const Interval< T_Bound, T_Rnd> & A_iv, 
                                             double A_x)
{
  return A_iv + T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator + (double A_x, 
                                             const Interval< T_Bound, T_Rnd> & A_iv)
{
  return A_iv + T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator - (const Interval< T_Bound, T_Rnd> & A_iv, 
                                             double A_x)
{
  return A_iv - T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator - (double A_x, 
                                             const Interval< T_Bound, T_Rnd> & A_iv)
{
  return T_Bound(A_x) - A_iv  ; 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator * (const Interval< T_Bound, T_Rnd> & A_iv, 
                                             double A_x)
{
  return A_iv * T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator * (double A_x, 
                                             const Interval< T_Bound, T_Rnd> & A_iv)
{
  return A_iv * T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator / (const Interval< T_Bound, T_Rnd> & A_iv, 
                                             double A_x)
{
  return A_iv / T_Bound(A_x); 
}

template < typename T_Bound, typename T_Rnd> 
inline Interval< T_Bound, T_Rnd> operator / (double A_x, 
                                             const Interval< T_Bound, T_Rnd> & A_iv)
{
  return  T_Bound(A_x) / A_iv ; 
}

*/

//////////////////////////////////////////////////////////////////////////
// These functions are called by arithmetic operators 
//    Operators are defined inside class to provide implicite conversions  
//    and therefore are 'inline' functions.
//    Because of that we only call the following non 'inline' functions
//////////////////////////////////////////////////////////////////////////

/// Interval + BoundType
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> add(const Interval< T_Bound, T_Rnd> & A_iv, 
                              const T_Bound &A_x);

/// Interval - BoundType
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> substract(const Interval< T_Bound, T_Rnd> & A_iv, 
                                    const T_Bound &A_x);
  
/// BoundType - Interval 
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> substract(const T_Bound  &A_x,
                                    const Interval< T_Bound, T_Rnd> & A_iv);

/// Interval * BoundType
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> multiply(const Interval< T_Bound, T_Rnd>& A_iv, 
                                   const T_Bound& A_x);

/// Interval / BoundType
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> divide(const Interval< T_Bound, T_Rnd>& A_iv, 
                                 const T_Bound& A_x);

///   BoundType / Interval 
template < typename T_Bound, typename T_Rnd> 
Interval< T_Bound, T_Rnd> divide(const T_Bound& A_x,
                                 const Interval< T_Bound, T_Rnd>& A_iv);


}} // namespace chomp::intervals 

#endif //_INTERVAL_FRIEND_H_

/// @}
