/// @addtogroup normalForms
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file linearSubstitution.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _linearSubstitution_HPP
#define _linearSubstitution_HPP

namespace chomp{
namespace normalForms{

// -------------------------------------------------------------------------- //

// the following procedure brings linear part to the normal form
// IMPORTANT: s should contain a partial derivatives, not the power series coefficients!
// on output: partial derivatives of 
//       N \circ s \circ M
// where M,N are linear - usually N=M^{-1}, however we do not specify here how the inverse is computed

template<typename MatrixType>
chomp::map::CnCoeff<MatrixType> linearSubstitution(
      const MatrixType& N,
      const chomp::map::CnCoeff<MatrixType>& s,
      const MatrixType& M
   )
{
   typedef chomp::vectalg::Multiindex::MultiindexVector MultiindexVector;
   typedef typename MatrixType::ScalarType ScalarType;
   int dimension = s.dimension();
   int rank = s.rank();

   chomp::vectalg::Multiindex::IndicesSet indices;
   chomp::vectalg::Multiindex::generateList(dimension,rank,indices);
   if(dimension!=M.numberOfRows() || dimension!=M.numberOfColumns() || dimension!=N.numberOfRows() || dimension!=N.numberOfColumns() )
   {
      throw std::runtime_error("normalForms::linearSubstitution - incompatible dimensions");
   }
   
   chomp::map::CnCoeff<MatrixType> result (dimension,rank);

   for(int r=1;r<=rank;++r)
   {
      chomp::vectalg::Multipointer mp = s.first(r);
      do{
         typename chomp::map::CnCoeff<MatrixType>::VectorType temp(dimension);
         MultiindexVector::iterator b=indices[r-1].begin(), e=indices[r-1].end();
         while(b!=e)
         {
            ScalarType product = M((*b)[0]+1,mp[0]+1);
            for(int j=1;j<r;++j)
               product *= M((*b)[j]+1,mp[j]+1);
            
            chomp::vectalg::Multipointer mp2(b->dimension(),b->begin());
            std::sort(mp2.begin(),mp2.end());
            temp += s(mp2)*product;
            ++b;
         }

         result(mp) = N*temp;
      }while(s.next(mp));
   }

   return result;
}


}} // namespace chomp::normalForms

#endif

/// @}
