/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file basicPoincareIntf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _BasicPoincareIntf_H
#define _BasicPoincareIntf_H

#include <string>

namespace chomp{
namespace poincare{

template<typename DS>
class BasicPoincareMap
{
public:
   typedef typename DS::MapType MapType;
   typedef typename DS::FunctionType FunctionType;
   typedef typename DS::MatrixType MatrixType;
   typedef typename DS::VectorType VectorType;
   typedef typename DS::ScalarType ScalarType;

   BasicPoincareMap(const MapType& dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

   BasicPoincareMap(const std::string& dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

   BasicPoincareMap(const char* dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

   VectorType operator()(VectorType);
   VectorType operator()(VectorType, MatrixType&);

   template <typename CnCoeffType>
   CnCoeffType& operator()(CnCoeffType& result);

// computations of derivatives of Poincare map from the serivatives of flow
   VectorType computeDT(const VectorType& Px, const MatrixType& derivativeOfFlow);
   MatrixType computeDP(const VectorType& Px, const MatrixType& derivativeOfFlow, const VectorType& dT);
   MatrixType computeD2T(
         const VectorType& Px,
         const MatrixType& derivativeOfFlow,
         const VectorType& DT,
         const MatrixType& DP,
         MatrixType hessianOfFlow[]
      );

   void computeD2P(
         const VectorType& Px,
         const MatrixType& derivativeOfFlow,
         const VectorType& DT,
         const MatrixType& DP,
         MatrixType hessianOfFlow[],
         const MatrixType& D2T,
         MatrixType result[]
      );

// the following function computes a partial derivatives of a Poincare map 
// to an arbitrarily order. On input - partial derivatives of the flow
   template<typename CnCoeffType>
   CnCoeffType computePartialDerivatives(const CnCoeffType& DPhi);

   const MapType& getField() const;
   const FunctionType& getSection() const;
   int getOrder() const;
   const ScalarType& getStep() const;

   void setField(const MapType& newField);
   void setSection(const FunctionType& newSection);
   void setOrder(int newOrder);
   void setStep(const ScalarType& newStep);

   void setFactor(double newFactor);
   void setParameter(const std::string& name, const ScalarType& value);

protected:
// the fields of the class

   DS dynamicalSystem; // a template dynamical system (Taylor, C2Taylor or CnTaylor)
   FunctionType section;
   MapType gradient;

// this value is used in the procedure reachSection.
// after this procedure the set is closer to section than sectionFactor * sizeOfSet
   double sectionFactor;
   int m_dim;
}; // end of template PoincareMap

// ------------------------------------------------------------------------------------------
template <typename DS>
template<typename CnCoeffType>
CnCoeffType BasicPoincareMap<DS>::computePartialDerivatives(const CnCoeffType& PhiSeries)
{
   typedef typename CnCoeffType::Multipointer Multipointer;

   const int rank = PhiSeries.rank();
   if(m_dim != PhiSeries.dimension())
      throw std::runtime_error("Incompatible dimensions in BasicPoincareMap::computePartialDerivatives");
   if(rank > this->dynamicalSystem.getRank())
      throw std::runtime_error("Incompatible orders in BasicPoincareMap::computePartialDerivatives");

   CnCoeffType PSeries(m_dim,rank); // result
   PSeries() = PhiSeries();
   if(rank<=0) return PSeries;

// fieldOnPx contains a series of vectorField computed at Px
   const typename DS::SeriesContainerType& DtPhiSeries = dynamicalSystem.computeDaPhi(PhiSeries,rank+1);

   VectorType gradientOnPx = gradient(PhiSeries());
   VectorType fieldOnPx =  getField()(PhiSeries());
   ScalarType scalarProduct = - gradientOnPx * fieldOnPx;

// first we compute separately dt/dx
   CnCoeffType TSeries(m_dim,rank);
   int i,j,k,s,r,p;
   for(j=0;j<m_dim;++j)
      for(i=0;i<m_dim;++i)
         TSeries(0,j) += gradientOnPx[i] * PhiSeries(i,j);
   for(j=0;j<m_dim;++j)
      TSeries(0,j) /= scalarProduct;

// and we compute dP/dx
   for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
         PSeries(i,j) = fieldOnPx[i]*TSeries(0,j) + PhiSeries(i,j);

// recursive computation of higher order partial derivatives
   for(p=2;p<=rank;++p)
   {
      Multipointer a = PSeries.first(p);
      do
      {
         VectorType temp = PhiSeries(a);
         VectorType temp2(m_dim);

         for(k=2;k<=p;++k)
         {
            const typename Multipointer::IndicesSet& is = Multipointer::generateList(p,k);
            typename Multipointer::IndicesSet::const_iterator b=is.begin(), e=is.end();
            while(b!=e)
            {
               ScalarType product=TSeries( 0,a.subMultipointer((*b)[0]) );
               for(j=1;j<k;++j)
                  product *= TSeries( 0,a.subMultipointer((*b)[j]) );

               DtPhiSeries.takeVector(Multipointer(0),k,temp2);
               temp += (product*factorial(k)) * temp2;

               for(s=0;s<k;++s)
               {
                  ScalarType product2(1.);
                  for(j=0;j<k;++j)
                     if(j!=s)
                        product2 *= TSeries( 0,a.subMultipointer((*b)[j]) );

                  DtPhiSeries.takeVector(a.subMultipointer((*b)[s]),k-1,temp2);
                  temp += product2*factorial(k-1)*temp2;
               }
               ++b;
            }
         }

         TSeries(0,a) = (gradientOnPx*temp)/scalarProduct;
         PSeries(a) = temp + TSeries(0,a)*fieldOnPx;
      }while(PSeries.next(a));
   }

   return PSeries;
}

// -----------------------------------------------------------------------------------------

template <typename DS>
template <typename CnCoeffType>
CnCoeffType& BasicPoincareMap<DS>::operator()(CnCoeffType& v)
// nonrigorous Poincare map
// the point just past section on the nonrigorous trajectory is returned
// the result containes an approximate values of the partial derivatives 
// of the Poincare map
{
   ScalarType sign;
   ScalarType step = dynamicalSystem.getStep();

   do{
      dynamicalSystem(v);
      sign = section(v());
   } while(sign == 0.);

//first approach to section with a constant - relatively large time step
   CnCoeffType beforeSection(m_dim,v.rank());
   while(section(v())*sign>0)
   {
      beforeSection = v;
      dynamicalSystem(v);
   }

// we approach section with small time steps as close as possible

   int order = dynamicalSystem.getOrder(), i;
   for(i=0;i<order;++i)
   {
      VectorType Grad = gradient(beforeSection());
      VectorType fieldDirection = dynamicalSystem.getField()(beforeSection());
      ScalarType approxDistance = abs(section(beforeSection()));
      ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
      ScalarType approxTime = 0.9*approxDistance/fieldSectionDirection;
      if(step<0)
         approxTime = -approxTime;
      dynamicalSystem.setStep(approxTime);
      dynamicalSystem(beforeSection);
      if (section(beforeSection())*sign<=0.) break;
   }

   while(section(beforeSection())*sign>0) dynamicalSystem(beforeSection);
   dynamicalSystem.setStep(step);
   v = beforeSection;
   return v;
}


// -----------------------  inline definitions ---------------------------------------------

template <typename DS>
BasicPoincareMap<DS>::BasicPoincareMap(
      const MapType& vField,
      const FunctionType& Section,
      int Order,
      const ScalarType& Step,
      double factor
   ) : dynamicalSystem(vField,Order,Step), section(Section)
{
   m_dim = vField.getDimension();
   gradient = section.gradient(m_dim);
   sectionFactor = factor;
}

// -----------------------------------------------------------------------------------------

template <typename DS>
BasicPoincareMap<DS>::BasicPoincareMap(
      const std::string& vField,
      const FunctionType& Section,
      int Order,
      const ScalarType& Step,
      double factor
   ) : dynamicalSystem(vField,Order,Step), section(Section)
{
   m_dim = dynamicalSystem.getField().getDimension();
   gradient = section.gradient(m_dim);
   sectionFactor = factor;
}

// -----------------------------------------------------------------------------------------

template <typename DS>
BasicPoincareMap<DS>::BasicPoincareMap(
      const char* vField,
      const FunctionType& Section,
      int Order,
      const ScalarType& Step,
      double factor
   ) : dynamicalSystem(vField,Order,Step), section(Section)
{
   m_dim = dynamicalSystem.getField().getDimension();
   gradient = section.gradient(m_dim);
   sectionFactor = factor;
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline const typename BasicPoincareMap<DS>::MapType& BasicPoincareMap<DS>::getField() const
{
   return dynamicalSystem.getField();
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline const typename BasicPoincareMap<DS>::FunctionType& BasicPoincareMap<DS>::getSection() const
{
   return section;
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline int BasicPoincareMap<DS>::getOrder() const
{
   return dynamicalSystem.getOrder();
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline const typename BasicPoincareMap<DS>::ScalarType& BasicPoincareMap<DS>::getStep() const
{
   return dynamicalSystem.getStep();
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setField(const MapType& newField)
{
   dynamicalSystem.setField(newField);
   m_dim = newField.getDimension();
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setOrder(int newOrder)
{
   dynamicalSystem.setOrder(newOrder);
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setStep(const ScalarType& newStep)
{
   dynamicalSystem.setStep(newStep);
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setFactor(double newFactor)
{
   sectionFactor = newFactor;
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setParameter(const std::string &name, const ScalarType& value)
{
   dynamicalSystem.setParameter(name,value);
}

// -----------------------------------------------------------------------------------------

template<typename DS>
inline void BasicPoincareMap<DS>::setSection(const FunctionType& newSection)
{
   section = newSection;
   gradient = section.gradient(m_dim);
}

}} // namespace chomp::poincare

#endif

/// @}
