/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file flpoinc.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _FLPOINC_HPP
#define _FLPOINC_HPP

#include "capd/poincare/basicPoincare.h"

namespace chomp{
namespace poincare{

// ---------------------- nonrigorous procedure ----------------------

template <typename DS>
typename BasicPoincareMap<DS>::VectorType BasicPoincareMap<DS>::operator()(VectorType v)
// nonrigorous Poincare map
// the point just past section on the nonrigorous trajectory is returned
{
   ScalarType sign;
   ScalarType step = dynamicalSystem.getStep();

   do{
      v = dynamicalSystem(v);
      sign = section(v);
   } while(sign == 0.);

//first approach to section with a constant - relatively large time step
   VectorType beforeSection(m_dim);
   while(section(v)*sign>0)
   {
      beforeSection = v;
      v = dynamicalSystem(v);
   }

// we approach section with small time steps as close as possible

   int order = dynamicalSystem.getOrder(), i;
   for(i=0;i<order;++i)
   {
      VectorType Grad = gradient(beforeSection);
      VectorType fieldDirection = dynamicalSystem.getField()(beforeSection);
      ScalarType approxDistance = abs(section(beforeSection));
      ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
      ScalarType approxTime = 0.9*approxDistance/fieldSectionDirection;
      if(step<0)
         approxTime = -approxTime;
      dynamicalSystem.setStep(approxTime);
      v = dynamicalSystem(beforeSection);
      if (section(v)*sign<=0.) break;
      beforeSection = v;
   }

   while(section(v)*sign>0) v = dynamicalSystem(v);
   dynamicalSystem.setStep(step);
   return v;
}

// ---------------------- nonrigorous procedure ----------------------

template <typename DS>
typename BasicPoincareMap<DS>::VectorType BasicPoincareMap<DS>::operator()(VectorType v, MatrixType &der)
// nonrigorous Poincare map
// the point just past section on the nonrigorous trajectory is returned
{
   ScalarType sign;
   ScalarType step = dynamicalSystem.getStep();
   der = MatrixType::Identity(m_dim);
   MatrixType oneStepDerivative(m_dim,m_dim);

   do{
      v = dynamicalSystem(v,oneStepDerivative);
      der = oneStepDerivative * der;
      sign = section(v);
   } while(sign == 0.);

//first approach to section with a constant - relatively large time step

   VectorType beforeSection(m_dim);
   MatrixType matrixBeforeSection(m_dim,m_dim);

   while(section(v)*sign>0)
   {
      beforeSection = v;
      matrixBeforeSection = der;
      v = dynamicalSystem(v,oneStepDerivative);
      der = oneStepDerivative * der;
   }

// we approach section with small time steps as close as possible

   int order = dynamicalSystem.getOrder(), i;
   for(i=0;i<order;++i)
   {
      VectorType Grad = gradient(beforeSection);
      VectorType fieldDirection = dynamicalSystem.getField()(beforeSection);
      ScalarType approxDistance = abs(section(beforeSection));
      ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
      ScalarType approxTime = 0.9*approxDistance/fieldSectionDirection;
      if(step<0)
         approxTime = -approxTime;
      dynamicalSystem.setStep(approxTime);
      v = dynamicalSystem(beforeSection,oneStepDerivative);
      der = oneStepDerivative * matrixBeforeSection;
      if (section(v)*sign<=0.) break;
      beforeSection = v;
      matrixBeforeSection = der;
   }

   while(section(v)*sign>0)
   {
      v = dynamicalSystem(v,oneStepDerivative);
      der = oneStepDerivative * der;
   }
   dynamicalSystem.setStep(step);
   return v;
}

}} // namespace chomp::poincare

#endif // #define _FLPOINC_HPP

/// @}
