/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file poincare.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _POINCARE_HPP
#define _POINCARE_HPP

namespace chomp{
namespace poincare{

template<typename DS> template<typename T>
typename PoincareMap<DS>::VectorType PoincareMap<DS>::operator()(T& originalSet)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template<typename DS> template<typename T>
typename PoincareMap<DS>::VectorType PoincareMap<DS>::operator()(T& originalSet, MatrixType& der)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = NULL;
   partialDerivatives = NULL;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template<typename DS> template<typename T>
typename PoincareMap<DS>::VectorType PoincareMap<DS>::operator()(T& originalSet, MatrixType& der, MatrixType hessian[])
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = &der;
   hessianOfFlow = hessian;
   partialDerivatives = NULL;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return crossSection<T>(originalSet);
}

/*__________________________________________________________________________*/

template<typename DS> template<typename T>
typename PoincareMap<DS>::VectorType PoincareMap<DS>::operator()(T& originalSet, CnCoeffType& result)
{
// the originalSet after this function contains a set just ater section
// we move the set close to section
   derivativeOfFlow = NULL;
   hessianOfFlow = NULL;
   partialDerivatives = &result;
   reachSection<T>(originalSet);

// and compute the PoincareMap image
   return result()=crossSection<T>(originalSet);
}

// -------------------- protected functions ---------------------------------

// this function is just for compatibility with C^1 and C^2 routines
// see the version for C^1-set types below
template<typename DS>
inline void PoincareMap<DS>::computeJacEnclosure(C0set& S, VectorType& V)
{
   S.move(dynamicalSystem);
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::computeJacEnclosure(C1set& S, VectorType& V)
{
   if(derivativeOfFlow!=NULL)
   {
      MatrixType enc = dynamicalSystem.jacEnclosure(V,*(S.getNorm()));
      MatrixType oneStepBound = enc*MatrixType(S);
      for(int i=0;i<m_dim;i++)
         for(int j=0;j<m_dim;j++)
            (*derivativeOfFlow)[i][j] = intervalHull((*derivativeOfFlow)[i][j],oneStepBound[i][j]);
   }
   S.move(dynamicalSystem);
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::computeJacEnclosure(C2set& S, VectorType& enc)
{
   if(derivativeOfFlow!=NULL)
   {
      int i,j,c;
      MatrixType jacEnc = dynamicalSystem.jacEnclosure(enc,*(S.getNorm()));
      MatrixType currentDerivative = MatrixType(S);

// computation of the hessian of Poincare map

      if(hessianOfFlow!=NULL)
      {
         MatrixType* oneStepHessian = new (m_dim,m_dim) MatrixType[m_dim];
         typename DS::C2CoeffType EH(m_dim); // EH will be an enclosure for the hessian part
         dynamicalSystem.c2Enclosure(enc,jacEnc,*(S.getNorm()),EH);
         int ss,r;

         for(i=0;i<m_dim;++i)
            oneStepHessian[i].clear();

         for(i=0;i<m_dim;++i)
            for(j=0;j<m_dim;++j)
               for(c=0;c<m_dim;++c)
                  for(ss=0;ss<m_dim;++ss)
                     for(r=0;r<m_dim;++r)
                        //oneStepHessian[i][j][c] += EH(i,ss,r) * (*derivativeOfFlow)[ss][j] * (*derivativeOfFlow)[r][c];
                        oneStepHessian[i][j][c] += EH(i,ss,r) * currentDerivative[ss][j] * currentDerivative[r][c];

         for(i=0;i<m_dim;++i)
            for(j=0;j<m_dim;++j)
               for(c=0;c<m_dim;++c)
                  for(ss=0;ss<m_dim;++ss)
                     oneStepHessian[i][j][c] += jacEnc[i][ss] * S(ss,j,c)/*hessianOfFlow[ss][j][c]*/;

         for(i=0;i<m_dim;i++)
            for(j=0;j<m_dim;j++)
               for(c=0;c<m_dim;c++)
                  hessianOfFlow[i][j][c] = intervalHull(hessianOfFlow[i][j][c],oneStepHessian[i][j][c]);

         delete []oneStepHessian;
      } // end if(hessianOfFlow!=NULL)

// computation of enclosure for derivative of the Poincare map
      MatrixType oneStepBound = jacEnc*currentDerivative;
      for(i=0;i<m_dim;i++)
         for(j=0;j<m_dim;j++)
            (*derivativeOfFlow)[i][j] = intervalHull((*derivativeOfFlow)[i][j],oneStepBound[i][j]);
   } // end of if(derivativeOfFlow!=NULL)

   S.move(dynamicalSystem);
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::computeJacEnclosure(CnSet& S, VectorType& )
{
   if(partialDerivatives!=NULL)
   {
      CnCoeffType enc(m_dim,partialDerivatives->rank()),
                  oneStepBound(m_dim,partialDerivatives->rank());

      dynamicalSystem.cnEnclosure(VectorType(S),enc,*(S.getNorm()));
      //dynamicalSystem.computeComposition(S.currentSet(),enc,oneStepBound);
      substitutionPowerSeries(enc,S.currentSet(),oneStepBound,false);

      for(int p=0;p<m_dim;++p)
      {
         typename CnCoeffType::iterator b=partialDerivatives->begin(p,1), e=partialDerivatives->end(p,partialDerivatives->rank());
         typename CnCoeffType::iterator i=oneStepBound.begin(p,1);
         while(b!=e)
         {
            *b = intervalHull(*b,*i);
            ++b;
            ++i;
         }
      }
   }
   S.move(dynamicalSystem);
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::saveMatrices(CnSet& S)
{
   if(partialDerivatives!=NULL) *partialDerivatives = S.currentSet();
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::saveMatrices(C2set& S)
{
   if(derivativeOfFlow!=NULL)
      *derivativeOfFlow = MatrixType(S);

   if(hessianOfFlow!=NULL)
   {
      int i,j,c;
      for(i=0;i<m_dim;++i)
         for(j=0;j<m_dim;++j)
            for(c=0;c<m_dim;++c)
               hessianOfFlow[i][j][c] = S(i,j,c);
   }
}

/*__________________________________________________________________________*/

template<typename DS>
void PoincareMap<DS>::saveMatrices(C1set& S)
{
   if(derivativeOfFlow!=NULL) *derivativeOfFlow = MatrixType(S);
}

/*__________________________________________________________________________*/

template <typename DS> template<typename T>
void PoincareMap<DS>::reachSection(T& theSet)
// The common procedure to reach section for all types of the sets.
// As a result we obtain the set closer to section than sectionFactor * sizeOfSet
{
   ScalarType step = dynamicalSystem.getStep();// time step used to reach the section
   ScalarType tmp;
   VectorType beforeSection;
   VectorType bound;
   int ok;

// first we verify if we can leave the section, i.e. if the inner product
// of the vectorFiled and the normal vector to the section is non-zero

   VectorType startSet = VectorType(theSet);
   while(section(startSet).contains(0.0))
      // one can find the poincare image of some set not contained
      // in the section. In that case we do not verify if the set can leave the section
   {
      VectorType startEnclosure = dynamicalSystem.enclosure(startSet,&ok);
      ScalarType innerProduct = gradient(startEnclosure)*dynamicalSystem.getField()(startEnclosure);
      if(innerProduct.contains(0.0))
      {
         throw PoincareException<T>(
            "Poincare section error: cannot leave the section",theSet,theSet,
            startEnclosure, dynamicalSystem.getField()(startEnclosure),
            section(startEnclosure), gradient(startEnclosure), innerProduct
         );
      }
      theSet.move(dynamicalSystem); // we make one step to leave the section and try again
      startSet = VectorType(theSet);
   }

// used for checking if we cross the section (if sign is changed)
   ScalarType check, sign;

// first return to section with constant time step
   beforeSection = startSet; // in fact it is equal to VectorType(theSet);
   check = section(beforeSection);
   sign = check.mid(); // we save the sign of the section function, check do not contain zero
   T temp(theSet);
   T *original = &theSet, *swapSet, *tempSet = &temp;

// now we try to move the set closer to section than one current time step
   while(check*sign>0)
   {
      bound = dynamicalSystem.enclosure(beforeSection,&ok);
      original->move(dynamicalSystem,*tempSet); // bound is enclosure for whole time step
      tmp = gradient(bound)*dynamicalSystem.getField()(bound);
      if (tmp.contains(0.0))
      {  // we check if we cross the section and than return during one step
         check = section(bound);
         if(subset(ScalarType(0.0),check))
         {
            throw PoincareException<T>(
               "Poincare section error:",theSet,temp,
               bound, dynamicalSystem.getField()(bound),
               check, gradient(bound), tmp
            );
         }
         beforeSection = VectorType(*tempSet);
      }
      else
      {
         beforeSection = VectorType(*tempSet);
         check = section(beforeSection);
      }
      swapSet = original;
      original = tempSet;
      tempSet = swapSet;
   }

   if(&theSet != tempSet) theSet = *tempSet;

// We want to come to section closer than  sectionFactor * size of set in section gradient direction

   beforeSection = VectorType(theSet);
   VectorType Grad = gradient(beforeSection);

   typename ScalarType::BoundType maxDistance = diam(abs(beforeSection*Grad)).rightBound()*sectionFactor;
   int maxIterations = 30; // usually 2-3 iterations are necessary to come close to section

   original = &theSet;
   tempSet = &temp;

   while(maxIterations)
   {
      VectorType Mid = midVector(beforeSection);
      VectorType fieldDirection = dynamicalSystem.getField()(Mid);
      ScalarType approxDistance = (abs(section(beforeSection))).leftBound();

      if(approxDistance.leftBound() < maxDistance)      // we are close enough
         break;

      ScalarType fieldSectionDirection = abs(fieldDirection * Grad);
      ScalarType approxTime = ((approxDistance-(maxDistance/15))/ fieldSectionDirection).mid();

      if (step<0)
         approxTime = -approxTime;

      dynamicalSystem.setStep(approxTime.leftBound());
      bound = dynamicalSystem.enclosure(beforeSection,&ok);
      tmp = gradient(bound)*dynamicalSystem.getField()(bound);
      original->move(dynamicalSystem,*tempSet);

      while(true)
      {
         if (tmp.contains(0.0)) // we check if we cross the section and than return during one step
         {
            check = section(bound);
            if(subset(ScalarType(0.),check))
            {
               throw PoincareException<T>(
                  "Poincare section gradient error:", theSet, temp,
                  bound, dynamicalSystem.getField()(bound),
                  check, gradient(bound), tmp
               );
            }
         }
         else
            check = section(VectorType(*tempSet));
         if(!(check*sign>0))
         {
            dynamicalSystem.setStep(dynamicalSystem.getStep() * TIME_STEP_CORECTION_FACTOR);
            bound = dynamicalSystem.enclosure(beforeSection,&ok);
            tmp = gradient(bound)*dynamicalSystem.getField()(bound);
            original->move(dynamicalSystem,*tempSet);
         }
         else break;
      }

      swapSet = original;
      original = tempSet;
      tempSet = swapSet;

      beforeSection = VectorType(*original);
      Grad = gradient(beforeSection);
      maxIterations--;
   }

   dynamicalSystem.setStep(step);
   if(&theSet != original) theSet = *original;
}

/*__________________________________________________________________________*/

// the following function is used to cross the section for all types of sets

template <typename DS> template<typename T>
typename PoincareMap<DS>::VectorType PoincareMap<DS>::crossSection(T& theSet)
{
// theSet is very close to the section
// we compute an approximate time needed to cross the section
// if the approximate values of time step is to small to change the sign of the section function
// the 'smallTime' time is used to move the set
   ScalarType step = dynamicalSystem.getStep();

   VectorType Set = VectorType(theSet);
   VectorType fieldDirection = dynamicalSystem.getField()(Set);
   VectorType normalVector = gradient(Set);
   ScalarType fieldSectionDirection = abs(fieldDirection * normalVector);
   ScalarType approxDistance = (abs(section(Set))).rightBound();

   ScalarType maxDistance = diam(abs(Set*normalVector)).rightBound();
   ScalarType smallTime = abs( maxDistance / (ScalarType(10.) * fieldSectionDirection).mid() );
   ScalarType approxTime = abs( approxDistance / (10.*fieldSectionDirection).mid());

   if(step<0)
   {
      approxTime = -approxTime;
      smallTime = -smallTime;
   }

   ScalarType check = section(Set);
   VectorType result = Set;
   typename ScalarType::BoundType sign = check.mid().rightBound(); // check does not contain zero, we save only the sign

   dynamicalSystem.setStep(approxTime);
   T temp(theSet); // we save the set before section, used in throwing exception only
   saveMatrices(theSet); // we save the \frac{d}{dx}\phi

// first attempt with approxTime step
   int ok;
   VectorType bound = dynamicalSystem.enclosure(Set,&ok);
   VectorType Fresult = dynamicalSystem.getField()(bound);
   if(subset(ScalarType(0.),gradient(bound)*Fresult)) // if the vector field is tangent to the section
   {
       VectorType secGradient = gradient(bound);
       throw PoincareException<T>(
          "Poincare section gradient error:", theSet, temp,
          bound, Fresult, check, secGradient, secGradient * Fresult
       );
   }

   computeJacEnclosure(theSet,bound); // in this procedure theSet.move(dynamicalSystem) is called
   Set = VectorType(theSet);
   for(int i=0;i<m_dim;i++)
   {
      if(!(Fresult[i].contains(0.)))
         result[i] = intervalHull(result[i],Set[i]);
      else
         result[i] = intervalHull(result[i],bound[i]); // bound is an enclosure
   }

// now we cross the section with very small step smallTime
   dynamicalSystem.setStep(smallTime);
   check = section(Set);

   while(!(check*sign<0)) // if the sign of the section has been changed we stop the procedure
   {
      bound = dynamicalSystem.enclosure(Set,&ok);
      Fresult = dynamicalSystem.getField()(bound);
      if(subset(ScalarType(0.),gradient(bound)*Fresult)) // if the vector field is tangent to the section
      {
         VectorType secGradient = gradient(bound);
         throw PoincareException<T>(
            "Poincare section gradient error:", theSet, temp,
            bound, Fresult, check, secGradient, secGradient * Fresult
         );
      }

// if the vector field is monotonic on the i-th coordinate
// the enclosure of the Poincare image is just the ScalarType hull
// of such coordinate on start and after one step
// i.e. phi_i([t,t+h],x) \subset intervalHull(phi_i(t,x),phi_i(t+h,x))
// in the other case we take the enclosure for whole trajectory
// i.e. phi_i([t,t+h],x)

      computeJacEnclosure(theSet,bound); // in this procedure theSet.move(dynamicalSystem) is called
      Set = VectorType(theSet);
      for(int i=0;i<m_dim;i++)
      {
         if(!(Fresult[i].contains(0.)))
            result[i] = intervalHull(result[i],Set[i]);
         else
            result[i] = intervalHull(result[i],bound[i]); // bound is an enclosure
      }

      check = section(Set); // we compute the sign of the section
   }

   dynamicalSystem.setStep(step);
   return result;
}

}} // namespace chomp::poincare

#endif // _POINCARE_HPP

/// @}
