/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file rowVector.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _RowVector_HPP
#define _RowVector_HPP

#include <stdexcept>
#include "capd/vectalg/rowVector.h"
#include "capd/vectalg/vectorintf.h"
#include "capd/vectalg/matrixintf.h"

template<typename Scalar,int cols>
Scalar RowVector<Scalar,cols>::euclNorm() const
{
   Scalar sum(0);
   const_iterator b=begin(), e=end();
   while(b!=e)
   {
      sum += static_cast<Scalar>(power(*b,2));
      ++b;
   }
   return Scalar(sqrt(nonnegativePart(sum)));
}

template<typename Scalar, int cols>
bool RowVector<Scalar,cols>::normalize()
{
   Scalar n;
   if( ! (( n=euclNorm()) > 0.0) )
   {
      return false;
   }else{
      iterator b=begin(), e=end();
      while(b!=e)
      {
         *b /=n;
         ++b;
      }
      return true;
   }
}

template<typename Scalar,int cols>
void RowVector<Scalar,cols>::clear()
{
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
}

template<typename Scalar, int cols>
RowVector<Scalar,cols>::operator Vector<Scalar,cols>() const
{
   Vector<Scalar,cols> result(dimension(),true);
   const_iterator b=begin(), e=end();
   typename Vector<Scalar,cols>::iterator i=result.begin();
   while(b!=e)
   {
      *i = *b;
      ++i;
      ++b;
   }
   return result;
}


template<typename Scalar,int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator=(const RowVector& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator= error - cannot change dimension of vector.");
   }
   const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i = *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar,int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator=(const Vector<Scalar,cols>& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator= error - cannot change dimension of vector.");
   }
   typename Vector<Scalar,cols>::const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i = *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar,int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator+=(const RowVector& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator+= error - different dimensions of vectors.");
   }
   const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i += *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar,int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator+=(const Vector<Scalar,cols>& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator+= error - different dimensions of vectors.");
   }
   typename Vector<Scalar,cols>::const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i += *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar,int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator-=(const RowVector& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator-= error - different dimensions of vectors.");
   }
   const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i -= *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar, int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator-=(const Vector<Scalar,cols>& a_rv)
{
   if(dimension()!=a_rv.dimension())
   {
      throw std::range_error("RowVector::operator-= error - different dimensions of vectors.");
   }
   typename Vector<Scalar,cols>::const_iterator b=a_rv.begin(), e=a_rv.end();
   iterator i = begin();
   while(b!=e)
   {
      *i -= *b;
      ++i;
      ++b;
   }
   return *this;
}

template<typename Scalar, int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator/=(const Scalar& a_s)
{
   typename RowVector<Scalar,cols>::iterator b=this->begin(), e=this->end();
   while(b!=e)
   {
      *b /= a_s;
      ++b;
   }
   return *this;
}

template<typename Scalar, int cols>
RowVector<Scalar,cols>& RowVector<Scalar,cols>::operator*=(const Scalar& a_s)
{
   typename RowVector<Scalar,cols>::iterator b=this->begin(), e=this->end();
   while(b!=e)
   {
      *b *= a_s;
      ++b;
   }
   return *this;
}

// ---------------------  operator+ ---------------------------------

template<typename Scalar, int cols>
Vector<Scalar,cols> operator+(const Vector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator+(Vector,RowVector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename Vector<Scalar,cols>::const_iterator b1 = A.begin();
   typename RowVector<Scalar,cols>::const_iterator b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 + *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar,int cols>
Vector<Scalar,cols> operator+(const RowVector<Scalar,cols>&A, const Vector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator+(RowVector,Vector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin();
   typename Vector<Scalar,cols>::const_iterator b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 + *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar, int cols>
Vector<Scalar,cols> operator+(const RowVector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator+(RowVector,RowVector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin(), b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 + *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}


// ---------------------  operator- ---------------------------------

template<typename Scalar,int cols>
Vector<Scalar,cols> operator-(const Vector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator-(Vector,RowVector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename Vector<Scalar,cols>::const_iterator b1 = A.begin();
   typename RowVector<Scalar,cols>::const_iterator b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 - *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar,int cols>
Vector<Scalar,cols> operator-(const RowVector<Scalar,cols>&A, const Vector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator-(RowVector,Vector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin();
   typename Vector<Scalar,cols>::const_iterator b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 - *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar,int cols>
Vector<Scalar,cols> operator-(const RowVector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator-(RowVector,RowVector) error - different dimensions of vectors.");
   }
   Vector<Scalar,cols> result(A.dimension(),true);
   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin(), b2 = B.begin();
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      *b = *b1 - *b2;
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

// ------------------------------- operaotr* -----------------------------------

template<typename Scalar,int cols>
Scalar operator*(const Vector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator*(Vector,RowVector) error - different dimensions of vectors.");
   }
   Scalar result(0);
   typename Vector<Scalar,cols>::const_iterator b1 = A.begin(), e1=A.end();
   typename RowVector<Scalar,cols>::const_iterator b2 = B.begin();
   while(b1!=e1)
   {
      result += (*b1) * (*b2);
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar, int cols>
Scalar operator*(const RowVector<Scalar,cols>&A, const Vector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator*(RowVector,Vector) error - different dimensions of vectors.");
   }
   Scalar result(0);
   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin(), e1=A.end();
   typename Vector<Scalar,cols>::const_iterator b2 = B.begin();
   while(b1!=e1)
   {
      result += (*b1) * (*b2);
      ++b1;
      ++b2;
   }
   return result;
}

template<typename Scalar,int cols>
Scalar operator*(const RowVector<Scalar,cols>&A, const RowVector<Scalar,cols>&B)
{
   if(A.dimension()!=B.dimension())
   {
      throw std::range_error ("operator*(RowVector,RowVector) error - different dimensions of vectors.");
   }
   Scalar result(0);

   typename RowVector<Scalar,cols>::const_iterator b1 = A.begin(), e1=A.end(), b2 = B.begin();
   while(b1!=e1)
   {
      result += (*b1) * (*b2);
      ++b1;
      ++b2;
   }

   return result;
}

// -----------------------------------------------------------------------------

template<typename Scalar,int rows,int cols>
Vector<Scalar,rows> operator*(const Matrix<Scalar,rows,cols>& a, const RowVector<Scalar,cols>&v)
{
   Vector<Scalar,rows> result(a.numberOfRows(),true);
   if(a.numberOfColumns()!=v.dimension())
      throw std::range_error("operator Matrix*RowVector: incompatible matrix dimensions");

   typename Vector<Scalar,rows>::iterator b=result.begin(), e=result.end();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i = a.cbegin();
   while(b!=e)
   {
      Scalar x(0);
      typename RowVector<Scalar,cols>::const_iterator bv=v.begin(), be=v.end();
      while(bv!=be)
      {
         x += (*bv) * (*i);
         ++bv;
         ++i;
      }
      *b=x;
      ++b;
   }
   return result;
}

template<typename Scalar,int cols>
Vector<Scalar,cols> operator-(const RowVector<Scalar,cols>&v)
{
   Vector<Scalar,cols> result(v.dimension(),true);
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   typename RowVector<Scalar,cols>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = -(*i);
      ++b;
      ++i;
   }
   return result;
}


// -----------------------------------------------------------------------------------

template<typename Scalar, int cols>
Vector<Scalar,cols> operator*(const Scalar& s, const RowVector<Scalar,cols>& v)
{
   Vector<Scalar,cols> result(v.dimension(),true);
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   typename RowVector<Scalar,cols>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = s * (*i);
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int cols>
Vector<Scalar,cols> operator*(const RowVector<Scalar,cols>& v, const Scalar& s)
{
   Vector<Scalar,cols> result(v.dimension(),true);
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   typename RowVector<Scalar,cols>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = s * (*i);
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int cols>
Vector<Scalar,cols> operator/(const RowVector<Scalar,cols>& v, const Scalar& s)
{
   Vector<Scalar,cols> result(v.dimension(),true);
   typename Vector<Scalar,cols>::iterator b=result.begin(), e=result.end();
   typename RowVector<Scalar,cols>::const_iterator i=v.begin();
   while(b!=e)
   {
      *b = (*i)/s;
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int cols>
std::ostream& operator<<(std::ostream& out, const RowVector<Scalar,cols>& v)
{
   out << "{";
   if(v.dimension()>0){
     //if(v[0]>=Scalar(0)) out << " "; /***** DW it does not work for complex vectors ***/
     out << v[0];
   }
   for(int i=1;i<v.dimension();i++)
   {
      out << ",";
      //if(v[i]>=Scalar(0)) out << " "; /***** DW it does not work for complex vectors ***/
      out << v[i];
   }
   out << "}";
   return out;
}

template<typename Scalar, int cols>
Vector<Scalar, cols> diam(const RowVector<Scalar, cols> & v)
{
   typedef typename RowVector<Scalar, cols>::ScalarType ScalarType;
   Vector<Scalar, cols> result(v.dimension());
   typename Vector<Scalar, cols>::iterator i=result.begin();
   typename RowVector<Scalar, cols>::const_iterator b=v.begin(), e=v.end();
   while(b!=e)
   {
      *i = (ScalarType(b->rightBound()) - ScalarType(b->leftBound())).rightBound();
      ++i;
      ++b;
   }
   return result;
}

#endif

/// @} 
