/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file vectorintf.h
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _vectorTemplate
#define _vectorTemplate

#include <iostream>
#include <cstdlib>
#include <vector>
#include <new>
#include "capd/capd/minmax.h"
#include "capd/capd/power.h"
#include "capd/vectalg/container.h"

inline double nonnegativePart(double x)
{
   return chomp::abs(x);
}

inline double nonnegativePart(int x)
{
   return chomp::abs(x);
}

template<typename Scalar, int dim>
class Vector;

template<typename Scalar, int dim>
Vector<Scalar,dim> abs(const Vector<Scalar,dim>& v);

template<typename Scalar, int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim>& v);

template<typename Scalar, int dim>
Vector<Scalar,dim> operator+(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
Scalar operator*(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
Vector<Scalar,dim> operator+(const Vector<Scalar,dim>& v, const Scalar& s);

template<typename Scalar, int dim>
Vector<Scalar,dim> operator-(const Vector<Scalar,dim>& v,const Scalar& s);

template<typename Scalar, int dim>
bool operator < (const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
bool operator > (const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
bool operator<=(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
bool operator>=(const Vector<Scalar,dim>& v1,const Vector<Scalar,dim>& v2);

template<typename Scalar, int dim>
std::ostream& operator<<(std::ostream& out, const Vector<Scalar,dim>& v);

template<typename Scalar, int dim>
std::istream& operator>>(std::istream& inp, Vector<Scalar,dim>& v);

template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator*(const Vector<Scalar,dim>& v, const FactorType& s);

template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator*(const FactorType& s,const Vector<Scalar,dim>& v);

template<typename Scalar, typename FactorType, int dim>
Vector<Scalar,dim> operator/(const Vector<Scalar,dim>& v, const FactorType& s);

// ---------------------------------------------------------------- //

template<typename IVector>
IVector intervalHull(const IVector&, const IVector&);

template<typename IVector>
void split(IVector&, IVector&);

template<typename IVector>
IVector midVector(const IVector&);

template<typename IVector>
IVector leftVector(const IVector&);

template<typename IVector>
IVector rightVector(const IVector&);

template<typename IVector>
typename IVector::ScalarType size(const IVector&);

template<typename IVector>
IVector diam(const IVector&);

template<typename IVector>
IVector intervalBall(const IVector&, const typename IVector::ScalarType &r);

template<typename IVector>
typename IVector::ScalarType solveAffineInclusion(const IVector&,const IVector&,const IVector&);

template<typename IVector>
typename IVector::ScalarType solveAffineInclusion(const IVector&,const IVector&,const IVector&,int&);

template<typename IVector>
bool subset(const IVector&, const IVector&);

template<typename IVector>
IVector intersection(const IVector&, const IVector&);

template<typename IVectorType>
bool intersection(const IVectorType &v1, const IVectorType &v2, IVectorType &result);


//########################### Vector template ################################//

template<typename Scalar, int dim>
class Vector : public chomp::vectalg::Container<Scalar,dim>
{
public:
   typedef Scalar ScalarType;
   typedef chomp::vectalg::Container<Scalar,dim> ContainerType;
   typedef typename ContainerType::iterator iterator;
   typedef typename ContainerType::const_iterator const_iterator;
   typedef Vector<Scalar,dim> VectorType;
   typedef Vector<int,dim> intVectorType;

   Vector(void){}
   explicit Vector(int A_dimension); // for compatibility with dynamic version
   Vector(const Scalar& x, const Scalar& y,const Scalar& z);  //obsolete, requires dimension=3
   Vector(int,ScalarType[]);
   Vector(const Vector&);
   explicit Vector(Vector<double,dim>&);
   Vector(int,bool); // it does not insert zeros
   void clear(void);                     //assign zero to each coord

   // assignments - vectors
   Vector& operator=  (const Vector& v);         //assign a vector
   Vector& operator+= (const Vector& v);         //increase by a vector
   Vector& operator-= (const Vector& v);         //decrease by a vector

   // assignments - Scalars
   Vector& operator=  (const Scalar& s);          //assign a Scalar
   Vector& operator+= (const Scalar& s);          //increase by a Scalar
   Vector& operator-= (const Scalar& s);          //decrease by a Scalar
   Vector& operator*= (const Scalar& s);          //rescale by multiplying
   Vector& operator/= (const Scalar& s);          //rescale by dividing

   using ContainerType::begin;
   using ContainerType::end;
//   using ContainerType::operator[];
//   using ContainerType::operator new[];
   using ContainerType::dimension;
   using ContainerType::resize;

   // Euclidean norm
   ScalarType euclNorm(void) const;
   //if possible vector is normalized and true is returned. Otherwise false is returned.
   bool normalize(void);
   void sorting_permutation(intVectorType& perm);
}; // end of class Vector template

// --------------------- inline definitions ------------------------------------ //

template<typename Scalar,int dim>
inline Vector<Scalar,dim>& Vector<Scalar,dim>::operator=(const Vector<Scalar,dim>& v)
{
   ContainerType::operator=(v);
   return *this;
}

template<typename Scalar,int dim>
inline Vector<Scalar,dim>::Vector(int A_dimension) : ContainerType(A_dimension)
{}

template<typename Scalar, int dim>
inline Vector<Scalar,dim>::Vector(const Vector& A_vect) : ContainerType(A_vect)
{}

template<typename Scalar,int dim>
inline Vector<Scalar,dim>::Vector(const Scalar& x,const Scalar& y,const Scalar& z) : ContainerType(3,true)
{
   (*this)[0]=x; (*this)[1]=y; (*this)[2]=z;
}

template<typename Scalar,int dim>
Vector<Scalar,dim>::Vector(int a_Dimension,bool) : ContainerType(a_Dimension,true)
{}

#endif // vectorTemplate

/// @}

