/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file cubisets.h
///
/// This file contains various procedures defined on cubical sets
/// and related to some combinatorial operations or homology computation.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: June 3, 2007.


#ifndef _CUBISETS_H_
#define _CUBISETS_H_

#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/homology/chains.h"
#include "chomp/struct/bitfield.h"
#include "chomp/struct/integer.h"
#include "chomp/struct/hashsets.h"
#include "chomp/homology/gcomplex.h"
#include "chomp/cubes/pointbas.h"
#include "chomp/cubes/cube.h"
#include "chomp/cubes/cell.h"
#include "chomp/homology/cubacycl.h"
#include "chomp/homology/tabulate.h"
#include "chomp/homology/known.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace chomp {
namespace homology {

// --------------------------------------------------
// ------------------- generators -------------------
// --------------------------------------------------

/// Writes projected homology generators of the cubical complex to a file.
/// Format: 0 = text format, 1 = chl format.
// Note: The main loop(s) of this function are copied from the function
// "createprojection".
template <class euclidom, class tCell>
std::ostream &writegenerators (std::ostream &out, const chain<euclidom> *hom,
	const chaincomplex<euclidom> &c, const gcomplex<tCell,euclidom> &g,
	const int *level, int xdim, int format = 0)
{
	bool firstlist = true;
	for (int d = 0; d <= c. dim (); d ++)
	{
		if ((level && !level [d]) || hom [d]. empty ())
			continue;
		if (firstlist)
			firstlist = false;
		else if (format == 0)
			out << '\n';
		if (format == 0)
		{
			if (hom [d]. size () == 1)
			{
				out << "The generator of H_" << d <<
					" follows:" << '\n';
			}
			else
			{
				out << "The " << hom [d]. size () <<
					" generators of H_" << d <<
					" follow:" << '\n';
			}
		}
		const hashedset<tCell> &cset = g [d];
		for (int i = 0; i < hom [d]. size (); i ++)
		{
			if (format == 0)
			{
				if (hom [d]. size () > 1)
					out << "generator " <<
						(i + 1) << '\n';
			}
			const chain<euclidom> &lst =
				c. gethomgen (d, hom [d]. num (i));
			for (int j = 0; j < lst. size (); j ++)
			{
				coordinate left [tCell::MaxDim];
				cset [lst. num (j)]. leftcoord (left);
				coordinate right [tCell::MaxDim];
				cset [lst. num (j)]. rightcoord (right);
				int projdim = 0;
				for (int k = 0; k < xdim; k ++)
				{
					if (left [k] != right [k])
						projdim ++;
				}
				if (projdim != d)
					continue;
				if (format == 0)
				{
					out << lst. coef (j) << " * " <<
						tCell (left, right, xdim) <<
						'\n';
				}
				else
				{
					for (int k = 0; k < xdim; k ++)
						out << left [k] << " ";
					for (int k = 0; k < xdim; k ++)
						out << right [k] << " ";
					out << lst. coef (j) << " " <<
						(i + 1) << " " << d << '\n';
				}
			}
		}
	}
	return out;
} /* writegenerators */


// --------------------------------------------------
// ------------------- map images -------------------
// --------------------------------------------------

/// Creates images of cells in 'm' as unions of cubes determined
/// by f1 and f2. The domains of interest must be given explicitly.
/// Returns the total number of cubes in all the images of cells.
template <class euclidom, class tCell, class tCube>
int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f1, const mvmap<tCube,tCube> &f2,
	const hashedset<tCube> &dom1, const hashedset<tCube> &dom2)
{
	int countimages = 0;
	coordinate leftbound [tCell::MaxDim];
	coordinate rightbound [tCell::MaxDim];
	coordinate left [tCell::MaxDim];
	coordinate right [tCell::MaxDim];
	for (int d = 0; d <= m. dim (); d ++)
	{
		const hashedset<tCell> &cset = m. get (d);
		if (cset. empty ())
			continue;
		const int spacedim = cset [0]. spacedim ();
		for (int i = 0; i < cset. size (); i ++)
		{
			cset [i]. leftcoord (left);
			cset [i]. rightcoord (right);
			for (int j = 0; j < spacedim; j ++)
			{
				// compensate for space wrapping
				if (right [j] < left [j])
					right [j] = left [j] + 1;
				// compute the bounds
				leftbound [j] = (coordinate) (left [j] -
					(left [j] == right [j]));
				rightbound [j] = (coordinate) (right [j] +
					(left [j] == right [j]));
			}
			rectangle r (leftbound, rightbound, spacedim);
			const coordinate *c;
			while ((c = r. get ()) != NULL)
			{
				if (!tCube::PointBase::check (c, spacedim))
					continue;
				tCube q (c, spacedim);
				if (dom1. check (q))
					m. add (d, i, f1 (q));
				if (dom2. check (q))
					m. add (d, i, f2 (q));
			}
			countimages += m (cset [i]). size ();
		}
	}
	return countimages;
} /* createimages */

/// A wrapper for the above function if there is only one map.
template <class euclidom, class tCell, class tCube>
inline int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f, const hashedset<tCube> &dom)
{
	mvmap<tCube,tCube> emptymap;
	hashedset<tCube> emptyset;
	return createimages (m, f, emptymap, dom, emptyset);
} /* createimages */

/// Creates images of cells in m as unions of cubes determined by f1 and f2.
/// Return the total number of cubes in all the images of cells.
template <class euclidom, class tCell, class tCube>
int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f1, const mvmap<tCube,tCube> &f2)
{
	const hashedset<tCube> &dom1 = f1. getdomain ();
	const hashedset<tCube> &dom2 = f2. getdomain ();
	return createimages (m, f1, f2, dom1, dom2);
} /* createimages */

/// A wrapper for the above function if there is only one map.
template <class euclidom, class tCell, class tCube>
inline int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f)
{
	mvmap<tCube,tCube> emptymap;
	return createimages (m, f, emptymap);
} /* createimages */


// --------------------------------------------------
// ------------------ projections -------------------
// --------------------------------------------------

/// Creates the chain map of the projection from a cell complex of the graph
/// of a map to a cell complex of the codomain of the map.
/// If a table of levels is given, creates the map only at given levels.
template <class euclidom, class tCell>
void createprojection (const gcomplex<tCell,euclidom> &Fcompl,
	const gcomplex<tCell,euclidom> &Ycompl,
	chainmap<euclidom> &cmap,
	int offset, int outdim, int discarddim, int *level = NULL)
{
	// go through the list of all the dimensions which are of concern
	for (int d = 0; d <= Ycompl. dim (); d ++)
	{
		if ((!level || level [d]) && (Fcompl. dim () >= d))
		{
			// take sets of cells of this dimension
			const hashedset<tCell> &Fset = Fcompl [d];
			if (Fset. empty ())
				continue;
			const hashedset<tCell> &Yset = Ycompl [d];
			if (Yset. empty ())
				continue;

			// go through the list of cells in Fcompl of dim. d
			for (int i = 0; i < Fset. size (); i ++)
			{
				// get this cell and its coordinates
				const tCell &Fcell = Fset [i];
				coordinate left [tCell::MaxDim];
				Fcell. leftcoord (left);
				coordinate right [tCell::MaxDim];
				Fcell. rightcoord (right);

				// check if this cell has no width in the
				// directions that are discarded
				register int j;
				for (j = 0; j < offset; j ++)
					if (left [j] != right [j])
					{
						j = offset + 33;
						break;
					}
				if (j > offset)
					continue;
				for (j = 0; j < discarddim; j ++)
					if (left [offset + outdim + j] !=
						right [offset + outdim + j])
					{
						j = discarddim + 33;
						break;
					}
				if (j > discarddim)
					continue;

				// create the projected cell
				if (!(tCell::PointBase::check
					(left + offset, outdim)))
					continue;
				if (!(tCell::PointBase::check
					(right + offset, outdim)))
					continue;
			//	tCell projected (left + offset,
			//		right + offset, outdim);
				tCell projected (Fcell, offset, outdim);

				// find its number in Y
				int nr = Yset. getnumber (projected);

				// if not found, discard it
				if (nr < 0)
					continue;

				// add the pair to the projection map
				euclidom e;
				e = 1;
				cmap. add (d, nr, i, e);
			}
		}
	}
	return;
} /* createprojection */

/// Creates the image of the projection from the set of cubical cells in the
/// given geometric complex to the subspace of R^n spanned by the 'outdim'
/// subsequent standard vectors with the first number 'offset'.
/// Only these cells which appear in 'only' are added unless 'only' is empty.
template <class euclidom, class tCell>
void project (const gcomplex<tCell,euclidom> &c,
	gcomplex<tCell,euclidom> &img,
	const gcomplex<tCell,euclidom> &only,
	int offset, int outdim, int discarddim, const int *level,
	bool watchforimages)
{
	// go through the list of all the dimensions which are of concern
	for (int d = 0; d <= c. dim (); d ++)
		if ((!level || level [d]) && (c. dim () >= d))
		{
			// take sets of cells of this dimension
			const hashedset<tCell> &cset = c [d];
			if (cset. empty ())
				continue;

			// go through the list of cells in c of dimension d
			for (int i = 0; i < cset. size (); i ++)
			{
				// get this cell and its coordinates
				const tCell &thecell = cset [i];
				coordinate left [tCell::MaxDim];
				thecell. leftcoord (left);
				coordinate right [tCell::MaxDim];
				thecell. rightcoord (right);

				// check if this cell has no width in the
				// directions that are discarded
				int j;
				for (j = 0; j < offset; j ++)
					if (left [j] != right [j])
					{
						j = offset + 33;
						break;
					}
				if (j > offset)
					continue;
				for (j = 0; j < discarddim; j ++)
					if (left [offset + outdim + j] !=
						right [offset + outdim + j])
					{
						j = discarddim + 33;
						break;
					}
				if (j > discarddim)
					continue;

				// add the projected cell to the complex
				if (!(tCell::PointBase::check
					(left + offset, outdim))
					|| !(tCell::PointBase::check
					(right + offset, outdim)))
				{
					if (watchforimages)
						throw "Inclusion undefined!";
					else
						continue;
				}
				tCell projected (left + offset,
					right + offset, outdim);
				if ((only. dim () >= 0) &&
					!only. check (projected))
					if (watchforimages)
						throw "Inclusion undefined.";
					else
						continue;
				img. add (projected);
			}
		}
	return;
} /* project */


// --------------------------------------------------
// ------------------- read cubes -------------------
// --------------------------------------------------

/// Reads a set of cubes from an input stream.
/// This is a specialization of the corresponding function
/// for reading general sets of elements.
std::istream &operator >> (std::istream &in, cubes &c);

/// Reads a cubical map from an input stream.
/// This is a specialization of the corresponding function
/// for reading a general multivalued map.
std::istream &operator >> (std::istream &in, cubicalmap &m);

// --------------------------------------------------

/// Reads the domain of a multivalued cubical map.
/// This is a specialization of the corresponding function
/// for reading the domain of a general multivalued map.
template <class tCube>
std::istream &readdomain (std::istream &in, hashedset<tCube> &dom)
{
	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		mvmap<tCube,tCube> Fdummy;
		return readdomain (in, dom, Fdummy);
	}

	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			tCube q;
			in >> q;
			if (!in)
				throw "Can't read the file.";
			dom. add (q);
		}
		ignoreline (in);
		ignorecomments (in);
	}
	
	return in;
} /* readdomain */

/// Reads the image of a multivalued cubical map.
/// This is a specialization of the corresponding function
/// for reading the image of a general multivalued map.
template <class tCube>
std::istream &readimage (std::istream &in, hashedset<tCube> &img)
{
	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		mvmap<tCube,tCube> Fdummy;
		return readimage (in, img, Fdummy);
	}

	typename tCube::CoordType left [tCube::MaxDim];
	typename tCube::CoordType right [tCube::MaxDim];
	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			tCube dummy;
			in >> dummy;
			in >> dummy;
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			in. get ();
			tCube q1, q2;
			in >> q1 >> q2;
			if (!in)
				throw "Can't read the file.";
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			int dim = q1. dim ();
			q1. coord (left);
			q2. coord (right);
			tRectangle<typename tCube::CoordType> r
				(left, right, dim);
			const typename tCube::CoordType *c;
			while ((c = r. get ()) != NULL)
				img. add (tCube (c, dim));
		}
		else
			ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readimage */

/// Read the image of a set under a multivalued cubical map.
/// This is a specialization of the corresponding function
/// for reading the image of a general multivalued map.
template <class tCube>
std::istream &readimage (std::istream &in, const hashedset<tCube> &dom,
	hashedset<tCube> &img)
{
	ignorecomments (in);
	// the general format: [x,y,z] -> {[a,b,c], [d,e,f]}
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		while (in. peek () != EOF)
		{
			if ((closingparenthesis (in. peek ()) == EOF) &&
				!std::isdigit (in. peek ()))
			{
				ignoreline (in);
				ignorecomments (in);
				continue;
			}
			tCube q;
			in >> q;
			bool ignore = !dom. check (q);
			ignorecomments (in);
			while (in. peek () == '-')
				in. get ();
			in. get (); // '>'
			ignorecomments (in);
			int opening = in. get ();
			int closing = closingparenthesis (opening);
			if (closing == EOF)
				throw "An opening brace '{' expected.";
			while ((in. peek () != EOF) &&
				(in. peek () != closing))
			{
				if (ignore)
				{
					if (in. get () == opening)
					{
						while ((in. peek () != EOF) &&
							(in. peek () != closing))
						{
							in. get ();
						}
						in. get (); // '}'
					}
				}
				else
				{
					in >> q;
					img. add (q);
					ignorecomments (in);
					if (in. peek () == ',')
					{
						in. get ();
						ignorecomments (in);
					}
				}
			}
		//	in. get (); // '}'
			ignoreline (in);
			ignorecomments (in);
		}
	}

	typename tCube::CoordType left [tCube::MaxDim];
	typename tCube::CoordType right [tCube::MaxDim];
	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			tCube domcube1, domcube2;
			in >> domcube1;
			if (!dom. check (domcube1))
			{
				ignoreline (in);
				ignorecomments (in);
				continue;
			}
			in >> domcube2;
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			in. get ();
			tCube q1, q2;
			in >> q1 >> q2;
			if (!in)
				throw "Can't read the file.";
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			if (dom. check (domcube1))
			{
				int dim = q1. dim ();
				q1. coord (left);
				q2. coord (right);
				tRectangle<typename tCube::CoordType> r
					(left, right, dim);
				const typename tCube::CoordType *c;
				while ((c = r. get ()) != NULL)
					img. add (tCube (c, dim));
			}
		}
		else
			ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readimage */

/// Reads the restriction of a multivalued map to the given pair of sets.
template <class tCube>
std::istream &readselective (std::istream &in, const hashedset<tCube> &dom1,
	const hashedset<tCube> &dom2, mvmap<tCube,tCube> &m)
{
	if (dom1. empty () && dom2. empty ())
	{
		sout << "Warning: The domain of the map is empty.\n";
		return in;
	}

	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
		return readselective (in, m, dom1, dom2);

	typename tCube::CoordType left [tCube::MaxDim];
	typename tCube::CoordType right [tCube::MaxDim];
	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			tCube domcube;
			in >> domcube;
			int dim = domcube. dim ();
			if (dom1. check (domcube) || dom2. check (domcube))
			{
				tCube q1, q2;
				in >> q1; // (ignored)
				ignorecomments (in);
				in. get ();
				ignorecomments (in);
				in. get ();
				in >> q1 >> q2;
				if (!in)
					throw "Can't read the file.";
				hashedset<tCube> &img = m [domcube];
				q1. coord (left);
				q2. coord (right);
				tRectangle<typename tCube::CoordType> r
					(left, right, dim);
				const typename tCube::CoordType *c;
				while ((c = r. get ()) != NULL)
					img. add (tCube (c, dim));
			}
		}
		ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readselective */

/// Reads a restriction of a multivalued cubical map to the given set.
/// The order of arguments is reversed to distinguish form the template
/// defined for a general multivalued map.
template <class tCube>
inline std::istream &readselective (std::istream &in,
	const hashedset<tCube> &dom, mvmap<tCube,tCube> &m)
{
	hashedset<tCube> empty;
	return readselective (in, dom, empty, m);
} /* readselective */


// --------------------------------------------------
// ------------------- neighbors --------------------
// --------------------------------------------------

/// Returns the maximal number of neighbors of a cube: 3^dim - 1.
int getmaxneighbors (int dim);

/// Returns the number of the neighbor bit for the given neighbor of 'q'
/// or -1 if not a neighbor or the same cube as 'q'.
template <class tCube>
int neighbor2bit (const tCube &q, const tCube &neighbor)
{
	coordinate c0 [tCube::MaxDim];
	q. coord (c0);
	coordinate c1 [tCube::MaxDim];
	neighbor. coord (c1);
	int dim = q. dim ();

	int number = 0;
	const coordinate *wrap = tCube::PointBase::getwrapping (dim);
	for (int i = 0; i < dim; i ++)
	{
		if (i)
			number *= 3;
		switch (c1 [i] - c0 [i])
		{
			case -1:
				number ++;
			case 1:
				number ++;
				break;
			case 0:
				break;
			default:
				if (!wrap || !wrap [i])
					return -1;
				if ((c1 [i] - c0 [i]) == (wrap [i] - 1))
					number += 2;
				else if ((c1 [i] - c0 [i]) == (1 - wrap [i]))
					number ++;
				else
					return -1;
				break;
		}
	}

	return number - 1;
} /* neighbor2bit */

/// Creates the neighbor of the given cube with the specified number.
/// If the coordinates of the neighbor do not belong to 'PointBase',
/// then returns 'q', unless 'unconditional' is set to true.
template <class tCube>
tCube bit2neighbor (const tCube &q, int number, bool unconditional = false)
{
	coordinate c0 [tCube::MaxDim];
	q. coord (c0);
	coordinate c1 [tCube::MaxDim];
	int dim = q. dim ();

	number ++;
	for (int i = dim - 1; i >= 0; i --)
	{
		switch (number % 3)
		{
			case 2:
				c1 [i] = c0 [i];
				c1 [i] --;
				break;
			case 1:
				c1 [i] = c0 [i];
				c1 [i] ++;
				break;
			case 0:
				c1 [i] = c0 [i];
				break;
			default:
				throw "Erratic neighbor.";
		}
		number /= 3;
	}

	if (unconditional || tCube::PointBase::check (c1, dim))
		return tCube (c1, dim);
	else
		return q;
} /* bit2neighbor */

/// Indicates the neighbors of the given cube from the given set in the
/// bit field. Returns the number of neighbors.
/// If the limit is nonzero, then quits after having found
/// that many neighbors.
template <class tCube, class tCubeSet>
int getneighbors (const tCube &q, BitField *bits, const tCubeSet &set,
	int limit = 0)
{
	// if the answer is trivial, return it
	if (set. empty ())
		return 0;

	// extract the dimension of the cube and an upper bound
	// for the number of its neighbors
	int dim = q. dim ();
	int maxneighbors = getmaxneighbors (dim);

	// would it be faster to go through the entire set?
	int usetheset = maxneighbors > set. size ();

	// prepare a counter for the number of neighbors
	int count = 0;

	// if the set is small
	if (usetheset)
	{
		// go through all the elements in the set
		for (int i = 0; i < set. size (); ++ i)
		{
			// if this is the current cube, ignore it
			if (set [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}
	else
	{
		// go through all possible neighbor numbers
		for (int number = 0; number < maxneighbors; number ++)
		{
			// if it is unknown whether the neighbor exists...
			if (!bits || !bits -> test (number))
			{
				// create a neighbor cube
				tCube neighbor = bit2neighbor (q, number);
				// if not existent, ignore it
				if (neighbor == q)
					continue;
				// if this cube is not in the set, ignore it
				if (!set. check (neighbor))
					continue;
				// set the appropriate bit in the bit field
				if (bits)
					bits -> set (number);
			}
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}

	return count;
} /* getneighbors */

/// Indicates the neighbors of the given cube from the union of the two
/// given sets in the bit field. Returns the number of neighbors.
/// If the limit is nonzero, then quits after having found
/// that many neighbors.
template <class tCube, class tCubeSet>
int getneighbors (const tCube &q, BitField *bits, const tCubeSet &set1,
	const tCubeSet &set2, int limit = 0)
{
	// if the answer is trivial, return it
	if (!set1. size () && !set2. size ())
		return 0;

	// extract the dimension of the cube and an upper bound
	// for the number of its neighbors
	int dim = q. dim ();
	int maxneighbors = getmaxneighbors (dim);

	// would it be faster to go through the entire set?
	int usetheset = maxneighbors > set1. size () + set2. size ();

	// prepare a counter for the number of neighbors
	int count = 0;

	// if the set is small
	if (usetheset)
	{
		// go through all the elements in the set
		for (int i = 0; i < set1. size (); i ++)
		{
			// if this is the current cube, ignore it
			if (set1 [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set1 [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
		// go through all the elements in the set
		for (int i = 0; i < set2. size (); i ++)
		{
			// if this is the current cube, ignore it
			if (set2 [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set2 [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}
	else
	{
		// go through all possible neighbor numbers
		for (int number = 0; number < maxneighbors; number ++)
		{
			// if it is unknown whether the neighbor exists...
			if (!bits || !bits -> test (number))
			{
				// create a neighbor cube
				tCube neighbor = bit2neighbor (q, number);
				// if not existent, ignore it
				if (neighbor == q)
					continue;
				// if this cube is not in the set, ignore it
				if (!set1. check (neighbor) &&
					!set2. check (neighbor))
					continue;
				// set the appropriate bit in the bit field
				if (bits)
					bits -> set (number);
			}
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}

	return count;
} /* getneighbors */

/// Adds neighbors listed in the given bit field to the set of cubes
/// unless they are in the 'forbidden' set.
template <class tCube, class tCubeSet>
int addneighbors (const tCube &q, const BitField &bits, tCubeSet &set,
	const tCubeSet &notthese)
{
	int maxneighbors = getmaxneighbors (q. dim ());

	int count = 0;
	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		tCube neighbor = bit2neighbor (q, number);
		if (!notthese. check (neighbor))
			set. add (neighbor);
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */

/// Adds neighbors listed in the given bit field to the set of cubes.
/// If not 'unconditional', then adds only cubes which already exist.
template <class tCube, class tCubeSet>
int addneighbors (const tCube &q, const BitField &bits, tCubeSet &set,
	bool unconditional = false)
{
	int maxneighbors = getmaxneighbors (q. dim ());
	int count = 0;

	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		tCube neighbor = bit2neighbor (q, number, unconditional);
		set. add (neighbor);
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */

/// Adds intersections of neighbors listed in the given bit field with the
/// given cube to the cubical complex.
/// If not 'unconditional', then adds only cubes which already exist.
template <class tCube, class tCell>
int addneighbors (const tCube &q, const BitField &bits,
	gcomplex<tCell,integer> &c, bool unconditional = false)
{
	int maxneighbors = getmaxneighbors (q. dim());

	int count = 0;
	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		tCube neighbor = bit2neighbor (q, number, unconditional);
		c. add (tCell (q, neighbor));
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */


// --------------------------------------------------
// ------------- neighborhoods for BDDs -------------
// --------------------------------------------------

/// The neighborhood of a cube in a set of cubes.
template <class cubetype, class settype>
class Neighbors
{
public:
	/// The default constructor.
	Neighbors (const cubetype &middle, const settype &cset,
		BitField &bfield): q (middle), s (cset), b (bfield) {};

	/// The procedure for checking whether the given neighbor exists.
	/// The number of the neighbor is consistent with the "bit2neighbor"
	/// and "neighbor2bit" procedures.
	bool check (int n) const
	{
		cubetype neighbor = bit2neighbor (q, n);
		if (neighbor == q)
			return false;
		b. set (n);
		return (s. check (neighbor));
	}

private:
	/// The cube whose neighbors are verified.
	const cubetype &q;

	/// The set of cubes in which the neighbors of the cube are sought.
	const settype &s;

	/// The bitfield to record each neighbor in.
	BitField &b;

}; /* class Neighbors */

/// The neighborhood of a cube in a union of two sets of cubes.
template <class cubetype, class settype>
class Neighbors2
{
public:
	/// The default constructor.
	Neighbors2 (const cubetype &middle, const settype &cset1,
		const settype &cset2, BitField &bfield):
		q (middle), s1 (cset1), s2 (cset2), b (bfield) {};

	/// The procedure for checking whether the given neighbor exists.
	/// The number of the neighbor is consistent with the "bit2neighbor"
	/// and "neighbor2bit" procedures.
	bool check (int n) const
	{
		cubetype neighbor = bit2neighbor (q, n);
		if (neighbor == q)
			return false;
		else return (s1. check (neighbor) || s2. check (neighbor));
	}

private:
	/// The cube whose neighbors are verified.
	const cubetype &q;

	/// The first set of cubes in which the neighbors
	/// of the cube are sought.
	const settype &s1;

	/// The first set of cubes in which the neighbors
	/// of the cube are sought.
	const settype &s2;

	/// The bitfield to record each neighbor in.
	BitField &b;

}; /* class Neighbors2 */


// --------------------------------------------------
// ------- acyclicity verification with BDDs --------
// --------------------------------------------------

/// The maximal dimension for which binary decision diagrams are programmed.
const int MaxBddDimPossible = 3;

/// The maximal dimension for which binary decision diagrams are used.
/// This can be decreased by a program if no binary decision diagrams
/// should be used. However, the value of MaxBddDim cannot exceed
/// the value of MaxBddDimPossible.
extern int MaxBddDim;

/// The variable which controls which binary decision diagrams
/// should be used in dimension 3, either programmed by P. Pilarczyk
/// (if set to false) or received from G. Malandain (if set to true).
extern bool UseMalandainCode;

// --------------------------------------------------

/// Uses binary decision diagrams to verify whether the neighborhood
/// of the given cube in the given set is acyclic. Uses the bitfield
/// provided to record each tested neighbor in.
/// Verifies whether the neighborhood of the given cube is acyclic or not
/// using the code generated from binary decision diagrams.
template <class tCube, class tCubeSet>
bool bddacyclic (const tCube &q, int dim, const tCubeSet &s,
	BitField &b)
{
	switch (dim)
	{
	case 1:
	{
		Neighbors<tCube,tCubeSet> n (q, s, b);
		return acyclic1d (n);
	//	break;
	}
	case 2:
	{
		Neighbors<tCube,tCubeSet> n (q, s, b);
		return acyclic2d (n);
	//	break;
	}
	case 3:
	{
		Neighbors<tCube,tCubeSet> n (q, s, b);
		if (UseMalandainCode)
			return acyclic3d_Malandain (n);
		else
			return acyclic3d (n);
	//	break;
	}
	}
	throw "Trying to use a BDD for dimension other than 1-3.";
} /* bddacyclic */

/// As above - for the union of the two sets.
template <class tCube, class tCubeSet>
bool bddacyclic (const tCube &q, int dim, const tCubeSet &s1,
	const tCubeSet &s2, BitField &b)
{
	switch (dim)
	{
	case 1:
	{
		Neighbors2<tCube,tCubeSet> n (q, s1, s2, b);
		return acyclic1d (n);
	//	break;
	}
	case 2:
	{
		Neighbors2<tCube,tCubeSet> n (q, s1, s2, b);
		return acyclic2d (n);
	//	break;
	}
	case 3:
	{
		Neighbors2<tCube,tCubeSet> n (q, s1, s2, b);
		if (UseMalandainCode)
			return acyclic3d_Malandain (n);
		else
			return acyclic3d (n);
	//	break;
	}
	}
	throw "Trying to use a BDD for dimension other than 1-3.";
} /* bddacyclic */


// --------------------------------------------------
// ------------ acyclicity verification -------------
// --------------------------------------------------

/// Checks whether this cube's nieghbors form an acyclic set.
/// Returns: 1 = yes for sure, 0 = probably no, but it is not sure.
int acyclic (int dim, BitField &b);

/// Checks thoroughly whether the given set of cubes is acyclic.
/// Returns true if yes, false if not.
template <class tCube>
inline bool acyclic (const hashedset<tCube> &cset)
{
	// the empty set is not acyclic
	if (cset. empty ())
		return false;

	// reduce the set and see if there is only one cube remaining
	hashedset<tCube> emptycubes;
	hashedset<tCube> theset = cset;
	cubreducequiet (theset, emptycubes, emptycubes);
	if (theset. size () == 1)
		return true;

	// create a cubical complex from this set of cubes
	gcomplex<typename tCube::CellType,integer> ccompl;
	int size = cset. size ();
	for (int i = 0; i < size; ++ i)
		ccompl. add (typename tCube::CellType (cset [i]));

	// check whether this geometric complex is acyclic or not
	return acyclic (ccompl);
} /* acyclic */

/// Verifies the acyclicity of a neighborhood of the given cube.
/// \param q - the cube whose neighborhood is verified
/// \param dim - the dimension of the cube
/// \param cset - the set to search for the cubes adjacent to <i>q</i>
/// \param b - a pointer to a bitfield in which neighbors are marked
/// \param maxneighbors - the maximal possible number of neighbors:
/// should be equal to 3^<i>dim</i> - 1
/// Uses BDDs if available. Marks verified neighbors in b.
template <class tCube>
inline bool acyclic (const tCube &q, int dim, const hashedset<tCube> &cset,
	BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
		return bddacyclic (q, dim, cset, *b);

	// get the neighbors of the cube
	int n = getneighbors (q, b, cset);

	// if the answer is obvious from the number of neighbors, return it
	if ((n == 0) || (n == maxneighbors))
		return false;
	if (n == 1)
		return true;

	// return the information on whether this set of neighbors is acyclic
	return acyclic (dim, *b);
} /* acyclic */

/// The same as above, but checking neighbors in the union of two sets.
template <class tCube>
inline bool acyclic (const tCube &q, int dim, const hashedset<tCube> &cset1,
	const hashedset<tCube> &cset2, BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
		return bddacyclic (q, dim, cset1, cset2, *b);

	// get the neighbors of the cube
	int n = getneighbors (q, b, cset1, cset2);

	// if the answer is obvious from the number of neighbors, return it
	if ((n == 0) || (n == maxneighbors))
		return false;
	if (n == 1)
		return true;

	// return the information on whether this set of neighbors is acyclic
	return acyclic (dim, *b);
} /* acyclic */

/// Verifies whether a cube from the other set can be removed.
template <class tCube>
bool acyclic_rel (const tCube &q, int dim, const hashedset<tCube> &cset,
	const hashedset<tCube> &other, BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
	{
		if (!getneighbors (q, b, cset, 1))
			return true;
		if (!bddacyclic (q, dim, other, *b))
			return false;
		return bddacyclic (q, dim, cset, other, *b);
	}

	// get the neighbors from the other set
	int nother = getneighbors (q, b, other);

	// verify if this set of neighbors is acyclic
	bool otheracyclic = (nother < maxneighbors) &&
		((nother == 1) || (nother && acyclic (dim, *b)));

	// add the neighbors from 'cset'
	int ncset = getneighbors (q, b, cset);

	// if there are no cubes in 'cset', then this cube is ok
	if (!ncset)
		return true;

	// if there are neighbors in 'cset' and the neighbors
	// from 'other' are not acyclic, skip this cube
	if (!otheracyclic)
		return false;

	// if there are neighbors from 'cset' and the neighbors
	// from both 'cset' and 'other' taken together
	// don't form an acyclic set, skip this cube
	if ((ncset + nother > 1) &&
		((ncset + nother == maxneighbors) ||
		!acyclic (dim, *b)))
		return false;
	return true;
} /* acyclic_rel */


// --------------------------------------------------
// ------------------ reduce cubes ------------------
// --------------------------------------------------

/// Computes the image of the face under the combinatorial cubical
/// multivalued map, but doesn't take the cube 'q' into consideration.
/// Returns the number of cubes whose images were added to 'img'.
template <class tCube, class tCell>
int computeimage (hashedset<tCube> &img, const tCell &face,
	const mvmap<tCube,tCube> &map, const hashedset<tCube> &cset,
	const tCube &ignore)
{
	// get the coordinates of the cubical cell
	typename tCell::CoordType left [tCell::MaxDim];
	face. leftcoord (left);
	typename tCell::CoordType right [tCell::MaxDim];
	face. rightcoord (right);

	// find the images of all the cubes containing this cell
	int spacedim = face. spacedim ();
	typename tCell::CoordType leftb [tCell::MaxDim];
	typename tCell::CoordType rightb [tCell::MaxDim];
	for (int j = 0; j < spacedim; j ++)
	{
		typename tCell::CoordType diff =
			(left [j] == right [j]) ? 1 : 0;
		leftb [j] = (left [j] - diff);
		rightb [j] = (right [j] + diff);
	}
	tRectangle<typename tCell::CoordType> r (leftb, rightb, spacedim);
	const typename tCell::CoordType *c;
	int countimages = 0;
	while ((c = r. get ()) != NULL)
	{
		if (!tCube::PointBase::check (c, spacedim))
			continue;
		tCube q (c, spacedim);
		if (q == ignore)
			continue;
		if (!cset. check (q))
			continue;
		img. add (map (q));
		countimages ++;
	}

	return countimages;
} /* computeimage */

/// Verifies if the map remains acyclic after the addition or removal of the
/// given cube to/from the union of the first and the second set.
/// Assumes that the map is acyclic before the change.
/// Returns 'true' if yes for sure, 'false' if there is any doubt about it.
template <class tCube>
bool remainsacyclic (const mvmap<tCube,tCube> &map, const tCube &q,
	const hashedset<tCube> &cset1, const hashedset<tCube> *cset2 = 0)
{
	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (q. dim ());

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// clear the neighborbits
	b. clearall (maxneighbors);

	// get the bitfield representing the set of the neighbors of the cube
	int n;
	if (cset2)
		n = getneighbors (q, &b, cset1, *cset2);
	else
		n = getneighbors (q, &b, cset1);

	// create all the faces of the cube
	gcomplex<typename tCube::CellType,integer> faces;
	addneighbors (q, b, faces);
	faces. addboundaries ();

	// compute the new images of all the faces
	// and determine if they are acyclic
	int startdim = faces. dim ();
	if (startdim >= n - 1)
		startdim --;
	for (int d = startdim; d >= 0; d --)
	{
		for (int i = 0; i < faces [d]. size (); i ++)
		{
			// compute the image of the face in the first set
			hashedset<tCube> img;
			int n = computeimage (img, faces [d] [i], map,
				cset1, q);

			// add the image of the second set if applicable
			if (cset2)
				n += computeimage (img, faces [d] [i], map,
					*cset2, q);

			// if this is the image of only one cube, it is Ok
			if (n == 1)
				continue;

			// verify that if the large image (with 'q')
			// can be reduced towards the small image
			hashedset<tCube> imgsurplus = map (q);
			imgsurplus. remove (img);
			cubreducequiet (img, imgsurplus);
			if (imgsurplus. size ())
				return false;
		}
	}

	return true;
} /* remainsacyclic */

/// Reduce the set 'cset' towards 'maincset'. These sets must be disjoint.
/// If 'quiet' is set to true, then suppresses any messages.
/// Returns the number of cubes removed from 'cset'.
template <class tCube>
int cubreducequiet (const hashedset<tCube> &maincset, hashedset<tCube> &cset,
	bool quiet = true)
{
	// remember the initial number of cubes in the set to be reduced
	int prev = cset. size ();

	// if the case is trivial, return the answer
	if (!prev)
		return 0;

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	hashedset<tCube> queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// remove cubes which can be removed
	// and add their neighbors to the queue
	for (int i = 0; i < cset. size (); i ++)
	{
		// show progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// take a cube to analyze
		const tCube &c = cset [i];

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, maincset, cset, &b, maxneighbors))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (c, &b, maincset, cset);
		
		// remove this cube from the queue
		queue. remove (c);

		// add its neighbors to the queue
		addneighbors (c, b, queue, maincset);

		// remove this cube from 'cset'
		if (!quiet)
			sseq << '0' << c << '\n';
		cset. removenum (i);
		i --;
	}

	// add a temporary progress indicator and reset the counter
	if (!quiet)
		scon << "*";
	count = 0;

	// take cubes from the queue and remove them if possible
	while (!queue. empty ())
	{
		// update the progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		tCube c = queue [0];
		queue. removenum (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, maincset, cset, &b, maxneighbors))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (c, &b, maincset, cset);

		// add the neighbors to the queue,
		// except for those who should be kept
		addneighbors (c, b, queue, maincset);

		// remove the cube from 'cset'
		if (!quiet)
			sseq << '0' << c << '\n';
		cset. remove (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'added'
	return prev - cset. size ();
} /* cubreducequiet */

/// Reduces a pair of sets of cubes for relative homology computation.
/// Returns the number of cubes removed from both sets.
template <class tCube>
inline int cubreduce (const hashedset<tCube> &maincset,
	hashedset<tCube> &cset)
{
	return cubreducequiet (maincset, cset, false);
} /* cubreduce */

/// Reduces a pair of sets of cubes for relative homology computation.
/// Does not remove any cubes from the set 'keep'. Additionally makes sure
/// that the acyclicity of the given map is preserved.
/// If 'quiet' is set to true, then suppresses any messages.
/// Returns the number of cubes removed from both sets.
template <class tCube>
int cubreducequiet (hashedset<tCube> &cset, hashedset<tCube> &other,
	mvmap<tCube,tCube> &map, const hashedset<tCube> &keep,
	bool quiet = true)
{
	// determine if the acyclicity of the map should be considered
	bool checkacyclic = !map. getdomain (). empty ();

	// remember the initial number of cubes in both sets
	int prev = cset. size () + other. size ();

	// if the case is trivial, return the answer
	if (cset. empty ())
	{
		if (!other. empty ())
		{
			hashedset<tCube> empty;
			other = empty;
		}
		return prev;
	}

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	hashedset<tCube> queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the other set, remove cubes which can be removed,
	// and add their neighbors to the queue
	int i;
	for (i = 0; i < other. size (); i ++)
	{
		// show the progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// if the cube should be kept, skip it
		if (keep. check (other [i]))
			continue;

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the cube cannot be removed, skip it
		if (!acyclic_rel (other [i], dim, cset, other, &b,
			maxneighbors))
		{
			continue;
		}

		// make sure that the acyclicity of the map on A is OK
		if (checkacyclic && !remainsacyclic (map, other [i], other))
			continue;

		// make sure that the acyclicity of the map on X is OK
		if (checkacyclic && !remainsacyclic (map, other [i],
			other, &cset))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (other [i], &b, cset, other);
		
		// remove this cube from the queue
		queue. remove (other [i]);

		// add its neighbors to the queue (except for 'keep')
		addneighbors (other [i], b, queue, keep);

		// remove this cube from 'other'
		if (!quiet)
			sseq << '0' << other [i] << '\n';
		other. removenum (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '.';
	count = 0;

	// go through the main set, remove cubes which can be removed,
	// and add their neighbors to the queue
	for (i = 0; i < cset. size (); i ++)
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// if this cube should be kept, skip it
		if (keep. check (cset [i]))
			continue;

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (cset [i], dim, cset, other, &b, maxneighbors))
			continue;

		// make sure that the acyclicity of the map on X is OK
		if (checkacyclic && !remainsacyclic (map, cset [i],
			cset, &other))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (cset [i], &b, cset, other);
		
		// remove this cube from the queue
		queue. remove (cset [i]);

		// add its neighbors to the queue
		addneighbors (cset [i], b, queue, keep);

		// remove this cube from 'cset'
		if (!quiet)
			sseq << '0' << cset [i] << '\n';
		cset. removenum (i);
		i --;
	}

	// update the temporary progress indicator and reset the counter
	if (!quiet)
		scon << "\b*";
	count = 0;

	// take cubes from the queue and remove them if possible
	while (!queue. empty ())
	{
		// update the progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue and remove it from the queue
		tCube c = queue [0];
		queue. removenum (0);

		// clean the neighborbits
		b. clearall (maxneighbors);

		// if this cube belongs to the other set...
		if (other. check (c))
		{
			if (!acyclic_rel (c, dim, cset, other, &b,
				maxneighbors))
			{
				continue;
			}

			// make sure that the map's acyclicity on A is OK
			if (checkacyclic && !remainsacyclic (map, c, other))
				continue;

			// make sure that the map's acyclicity on X is OK
			if (checkacyclic && !remainsacyclic (map, c,
				other, &cset))
				continue;

			// fill in the bit field if it has not been done yet
			if (dim <= MaxBddDim)
				getneighbors (c, &b, cset, other);
		
			// add the neighbors of this cube to the
			// queue, but don't add neighbors from 'keep'
			addneighbors (c, b, queue, keep);

			// remove the cube from the other set
			if (!quiet)
				sseq << '0' << c << '\n';
			other. remove (c);
		}

		// otherwise, if this cube belongs to 'cset'...
		else
		{
			// if the neighborhood of the cube is not acyclic, skip it
			if (!acyclic (c, dim, cset, other, &b, maxneighbors))
				continue;

			// fill in the bit field if it has not been done yet
			if (dim <= MaxBddDim)
				getneighbors (c, &b, cset, other);

			// add the neighbors to the queue,
			// except for those who should be kept
			addneighbors (c, b, queue, keep);

			// remove the cube from 'cset'
			if (!quiet)
				sseq << '0' << c << '\n';
			cset. remove (c);
		}
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from both sets
	return prev - cset. size () - other. size ();
} /* cubreducequiet */

/// Reduces a pair of sets of cubes for relative homology computation.
/// Does not remove any cubes from the set 'keep'. Additionally makes sure
/// that the acyclicity of the given map is preserved.
/// Returns the number of cubes removed from both sets.
template <class tCube>
inline int cubreduce (hashedset<tCube> &cset, hashedset<tCube> &other,
	mvmap<tCube,tCube> &map, const hashedset<tCube> &keep)
{
	return cubreducequiet (cset, other, map, keep, false);
} /* cubreduce */

/// Reduces a pair of sets of cubes for relative homology computation.
/// Does not remove any cubes from the set 'keep'.
/// If 'quiet' is set to true, then suppresses any messages.
/// Returns the number of cubes removed from both sets.
template <class tCube>
inline int cubreducequiet (hashedset<tCube> &cset, hashedset<tCube> &other,
	const hashedset<tCube> &keep, bool quiet = true)
{
	mvmap<tCube,tCube> emptymap;
	return cubreducequiet (cset, other, emptymap, keep, quiet);
} /* cubreducequiet */

/// Reduces a pair of sets of cubes for relative homology computation.
/// Does not remove any cubes from the set 'keep'.
/// Returns the number of cubes removed from both sets.
template <class tCube>
inline int cubreduce (hashedset<tCube> &cset, hashedset<tCube> &other,
	const hashedset<tCube> &keep)
{
	return cubreducequiet (cset, other, keep, false);
} /* cubreduce */


// --------------------------------------------------
// ------------------ expand cubes ------------------
// --------------------------------------------------

/// Expands the set 'other' towards 'cset' without changing the homology
/// of (cset + other, other). The two sets must be disjoint.
template <class tCube>
int cubexpand (hashedset<tCube> &cset, hashedset<tCube> &other,
	bool quiet = false)
{
	// if the case is trivial, return the answer
	if (cset. empty () || other. empty ())
		return 0;

	// remember the initial number of cubes in the main set
	int prev = cset. size ();

	// determine the dimension of the cubes
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	hashedset<tCube> queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the main set, move cubes to the other set
	// if possible, and add their neighbors to the queue
	for (int i = 0; i < cset. size (); i ++)
	{
		// show progress indicator if suitable
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (cset [i], dim, other, &b, maxneighbors))
			continue;
		
		// remove this cube from the queue
		queue. remove (cset [i]);

		// add neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (cset [i], &b, cset);
		addneighbors (cset [i], b, queue);

		// move the cube to the other set
		if (!quiet)
			sseq << '2' << cset [i] << '\n';
		other. add (cset [i]);
		cset. removenum (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '*';
	count = 0;

	// take cubes from the queue and move them from 'cset' to 'other'
	while (queue. size ())
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		const tCube c = queue [0];
		queue. removenum (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;

		// add the neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move this cube from 'cset' to 'other'
		if (!quiet)
			sseq << '2' << c << '\n';
		cset. remove (c);
		other. add (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'cset'
	return prev - cset. size ();
} /* cubexpand */

/// Expands the set 'other' towards 'cset' without changing the homology
/// of (cset + other, other). The two sets must be disjoint.
/// Increases 'img' if necessary to cover F (other),
/// but moves cubes only if it can prove that the inclusion of the old 'img'
/// into the new, enlarged 'img' induces an isomorphism in homology.
/// Every cube added to 'img' is removed from 'imgsrc'.
template <class tCube>
int cubexpand (hashedset<tCube> &cset, hashedset<tCube> &other,
	hashedset<tCube> &imgsrc, hashedset<tCube> &img,
	const mvmap<tCube,tCube> &map, bool indexmap,
	bool checkacyclic, bool quiet = false)
{
	// if the case is trivial, return the answer
	if (cset. empty () || other. empty ())
		return 0;

	// remember the initial number of cubes in the main set
	int prev = cset. size ();

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	hashedset<tCube> queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the main set, move cubes to the other set
	// if possible, and add their neighbors to the queue
	for (int i = 0; i < cset. size (); i ++)
	{
		// show progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// take a cube from 'cset'
		const tCube &c = cset [i];

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;
		
		// take the image and reduce what is outside 'img'
		const hashedset<tCube> &cubimage = map (c);
		hashedset<tCube> ximage = cubimage;
		if (indexmap)
			ximage. add (c);
		ximage. remove (img);
		cubreducequiet (img, ximage);

		// if the image could not be reduced, skip the cube
		if (!ximage. empty ())
			continue;

		// make sure that the acyclicity of the map is not spoiled
		if (checkacyclic && !remainsacyclic (map, c, other))
			continue;

		// add the image of this cube to the image of the map
		if (!quiet && !cubimage. empty ())
		{
			sseq << "R\n";
			for (int j = 0; j < cubimage. size (); j ++)
				sseq << '2' << cubimage [j] << '\n';
			if (indexmap)
				sseq << '2' << c << '\n';
		}
		img. add (cubimage);
		if (indexmap)
			img. add (c);

		// remove from 'imgsrc' all the cubes added to 'img'
		imgsrc. remove (cubimage);
		if (indexmap)
			imgsrc. remove (c);

		// remove this cube from the queue
		queue. remove (c);

		// add neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move the cube to the other set
		if (!quiet)
			sseq << "L\n" << '2' << c << '\n';
		other. add (c);
		cset. removenum (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '*';
	count = 0;

	// take cubes from the queue and move them from 'cset' to 'other'
	while (!queue. empty ())
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		tCube c = queue [0];
		queue. removenum (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;
		
		// take the image and reduce what is outside 'img'
		const hashedset<tCube> &cubimage = map (c);
		hashedset<tCube> ximage = cubimage;
		if (indexmap)
			ximage. add (c);
		ximage. remove (img);
		cubreducequiet (img, ximage);

		// if the image could not be reduced, skip the cube
		if (!ximage. empty ())
			continue;

		// make sure that the acyclicity of the map is not spoiled
		if (checkacyclic && !remainsacyclic (map, c, other))
			continue;

		// add the image of this cube to the image of the map
		if (!quiet && !cubimage. empty ())
		{
			sseq << "R\n";
			for (int i = 0; i < cubimage. size (); i ++)
				sseq << '2' << cubimage [i] << '\n';
			if (indexmap)
				sseq << '2' << c << '\n';
		}
		img. add (cubimage);
		if (indexmap)
			img. add (c);

		// remove from 'imgsrc' all the cubes added to 'img'
		imgsrc. remove (cubimage);
		if (indexmap)
			imgsrc. remove (c);

		// add the neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move this cube from 'cset' to 'other'
		if (!quiet)
			sseq << "L\n" << '2' << c << '\n';
		cset. remove (c);
		other. add (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'cset'
	return prev - cset. size ();
} /* cubexpand */


} // namespace homology
} // namespace chomp

#endif // _CUBISETS_H_

/// @}

