/// @addtogroup multiwork
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file coord.h
///
/// This file contains the definition of the MultiWork coordinator class.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on August 11, 2004. Last revision: January 13, 2006.


#ifndef _MWCOORD_H_
#define _MWCOORD_H_

#include "chomp/multiwork/task.h"


namespace chomp {
namespace multiwork {

class mwWorker;


// --------------------------------------------------
// ----------------- mwCoordinator ------------------
// --------------------------------------------------

/// This class defines a generic coordinator task object for the multi-work
/// distributed computations framework.
/// Each coordinator class defined by the user must inherit from this class.
class mwCoordinator: public virtual mwTask
{
public:
	// --- constructor/destructor ---
	
	// the default constructor
	mwCoordinator ();
		
	// the destructor
	virtual ~mwCoordinator ();

	// --- configuration ---

	// keep the workers running after the coordinator's completion
	void KeepWorkers (bool keep = true);

	// save addresses of workers to a file; replace the filename's
	// extension with the given one (works for Unix and Windows names);
	// repeated address:port combinations are saved once
	int Save (const char *filename, const char *extension = NULL);

	// --- run the coordinator ---

	// run the coordinator until all the work has been completed;
	// if a worker's address is supplied, the work will be done locally
	// unless any remote worker is connected to the coordinator;
	// return mwOk or mwError
	int Coordinate (mwWorker *w = NULL);

	// --- the coordinator's functions ---

	// send a portion of init data to be processed;
	// return mwOk on success, mwError if the buffer is full
	// and the data was ignored
	int Init (mwData &data);

	// send a portion of data to be processed;
	// return mwOk on success, mwError if the data buffer is full
	// and the data should be sent again after having received
	// at least one processed data portion with the 'Receive' function
	int Send (mwData &data);

	// receive a portion of processed data (non-blocking);
	// return mwOk on success, mwError - no data available yet
	int Receive (mwData &data);
		
	// retrieve the data rejected by one of the workers;
	// return mwOk on success, mwError if there is no such data
	int Recall (mwData &data);
		
	// run the multi-work main procedure: send the data to tasks
	// and receive the results of their computations;
	// options: mwOk (default), mwNoMoreData (finalize tasks);
	// return (test bits): mwArrived - data arrived
	// (use Receive to retrieve it), mwHungry - data is needed
	// (use Send to supply the data for processing),
	// mwRejected - a worker rejected some data (use Recall
	// to retrieve this data), mwCompleted - tasks completed
	// (some data that arrived may still need to be retrieved)
	int Run (int options = mwNone);

	// disconnect all the workers (normally called in the destructor)
	void Disconnect ();

private:
	// --- coordinator's job ---

	// prepare a piece of data to be sent to a worker;
	// return mwOk on success, mwNothing if there is no data,
	// mwError to cancel the computations;
	// overload this function in the class you derive from mwCoordinator
	virtual int Prepare (mwData &data);
	
	// accept a result received from a worker;
	// return mwOk on success, mwError to cancel the computations;
	// overload this function in the class you derive from mwCoordinator
	virtual int Accept (mwData &data);
	
	// acknowledge data that was rejected by a worker;
	// return mwOk on success, mwError to cancel the computations;
	// overload this function in the class you derive from mwCoordinator
	virtual int Reject (mwData &data);

	// --- configuration ---

	// should data be processed locally only?
	bool singleWork;
	
	// the address of a local worker or NULL if none
	mwWorker *localWorker;

	// should the workers be kept running after coordinator is done?
	bool keepWorkers;

	// has some initialization data caused a fatal error locally?
	bool badInitialization;

	// the number of initialization data pieces processed locally
	int singleinitsent;
	
	// tables of mwData pieces to be sent to tasks
	mwData xInit [mwMAXINIT], xRejected [mwMAXDATA];
	mwData xWork [mwMAXDATA], xDone [mwMAXDATA];
	int nInit, nRejected, nWork, nDone;

	// --- data used for the communication with workers ---

	// is this the first run of the 'run' procedure?
	bool firstrun;

	// is all the data sent to workers and the coordinator is waiting?
	bool alldatasent;

	// socket at which new workers are listened to
	int listensocket;

	// the total number of workers (some of them can be disconnected)
	int nworkers;

	// the number of workers that are actually connected
	int nworking;

	// the socket numbers for the communication with workers
	// (negative if the worker is disconnected)
	int workers [mwMAXWORK];

	// the computer name of each worker
	mwData names [mwMAXWORK];

	// the ports at which the workers are going to listen next time
	int ports [mwMAXWORK];

	// the number of initialization data pieces sent to each worker
	int sentinit [mwMAXWORK];

	// the data sent to each worker (-1 = none, -2 = ignore result)
	int sentdata [mwMAXWORK];

	// the i/o flags of the workers + one flag for 'listensocket'
	int ioflags [mwMAXWORK + 1];

}; /* class mwCoordinator */

// --------------------------------------------------

inline mwCoordinator::mwCoordinator ():
#if mwNETWORK
	singleWork (false),
#else
	singleWork (true),
#endif
	localWorker (NULL),
	keepWorkers (false),
	badInitialization (false),
	singleinitsent (0),
	nInit (0), nRejected (0), nWork (0), nDone (0),
	firstrun (true),
	alldatasent (true),
	listensocket (-1),
	nworkers (0),
	nworking (0)
{
	return;
} /* mwCoordinator::mwCoordinator */

inline mwCoordinator::~mwCoordinator ()
{
	// disconnect all the workers
	Disconnect ();
	
	return;
} /* mwCoordinator::~mwCoordinator */

inline void mwCoordinator::KeepWorkers (bool keep)
{
	keepWorkers = keep;
	return;
} /* mwCoordinator::KeepWorkers */

inline int mwCoordinator::Save (const char *filename, const char *extension)
{
	// don't erase the file if there are no workers
	if (!nworkers)
		return mwOk;

	char *name = SetExtension (filename, extension);
	std::ofstream f (name);
	delete [] name;
	if (!f)
		return mwError;

	// list the connected workers first (conn == 0),
	// and disconnected workers later (conn == 1)
	for (int conn = 0; conn < 2; conn ++)
	{
		// this is the first time an address is listed in its group
		bool firsttime = true;

		// go through the entire list of workers
		for (int i = 0; i < nworkers; i ++)
		{
			// determine the host name
			const char *host = (const char *) (names [i]);

			// if no host name or port number is known, skip it
			if (!host || !*host || (ports [i] <= 0))
				continue;

			// if this is a worker from the other group, skip it
			if (!conn && (workers [i] < 0))
				continue;
			if (conn && (workers [i] >= 0))
				continue;

			// check if the same worker was already listed
			int j;
			for (j = 0; j < nworkers; ++ j)
			{
				// if 'i' is connected and there was no
				// identical worker before 'i', list it!
				if (!conn && (j >= i))
					break;
				// do not compare with self
				if (j == i)
					continue;
				// if 'i' is connected, but the other is not,
				// skip the other worker
				if (!conn && (workers [j] < 0))
					continue;
				// if the ports are different, skip worker j
				if (ports [i] != ports [j])
					continue;
				// if the names are different, skip worker j
				const char *hst = (const char *) (names [j]);
				if (!hst || !*hst)
					continue;
				if (strcmp (host, hst))
					continue;
				// if both are connected, or the other one
				// is connected, or j < i, don't list it!
				if (!conn || (workers [j] >= 0) || (j < i))
					j = nworkers + 1;
			}
			if (j > nworkers)
				continue;

			// if listed for the first time, write the group name
			if (firsttime)
			{
				if (!conn)
					f << "; Connected workers:\n";
				else
					f << "; Not connected:\n";
				firsttime = false;
			}

			// write the address
			f << host << ":" << ports [i] << '\n';
		}
	}
	
	if (!keepWorkers)
		f << "; The workers exit upon disconnection.\n";
	else
		f << "; The workers remain running after disconnection.\n";

	if (!f)
		return mwError;
	return mwOk;
} /* mwCoordinator::Save */

// --------------------------------------------------

inline int mwCoordinator::Prepare (mwData &)
{
	return mwNoData;
} /* mwCoordinator::Prepare */

inline int mwCoordinator::Accept (mwData &)
{
	return mwOk;
} /* mwCoordinator::Accept */

inline int mwCoordinator::Reject (mwData &)
{
	return mwOk;
} /* mwCoordinator::Reject */

// --------------------------------------------------

inline int mwCoordinator::Init (mwData &x)
{
	if (nInit >= mwMAXINIT)
		return mwError;
	xInit [nInit]. Take (x);
	xInit [nInit ++]. Rewind ();
	alldatasent = false;
	return mwOk;
} /* mwCoordinator::Init */

inline int mwCoordinator::Send (mwData &x)
{
	if (nWork >= mwMAXDATA)
		return mwError;
	xWork [nWork]. Take (x);
	xWork [nWork ++]. Rewind ();
	alldatasent = false;
	return mwOk;
} /* mwCoordinator::Send */

inline int mwCoordinator::Receive (mwData &x)
{
	if (!nDone)
		return mwError;
	x. Take (xDone [-- nDone]);
	x. Rewind ();
	return mwOk;
} /* mwCoordinator::Receive */

inline int mwCoordinator::Recall (mwData &x)
{
	if (!nRejected)
		return mwError;
	x. Take (xRejected [-- nRejected]);
	x. Rewind ();
	return mwOk;
} /* mwCoordinator::Recall */

inline int mwCoordinator::Coordinate (mwWorker *w)
{
	// remember the local worker's address
	localWorker = w;
	if (localWorker)
		localWorker -> localCoord = this;

	// data that is waiting to be sent to a worker
	mwData current;
	bool data_waiting = false;

	while (1)
	{
		// prepare data to wait for being sent a worker
		if (!data_waiting)
		{
			current. Reset ();
			switch (Prepare (current))
			{
			case mwOk:
				data_waiting = true;
				break;
			case mwError:
				return mwError;
			case mwNoData:
				break;
			}
		}

		// run the task manager
		int flags = data_waiting ? mwNone : mwNoMoreData;
		flags = Run (flags);
		
		// if some new data arrived, process it
		if (flags & mwArrived)
		{
			mwData data;
			Receive (data);
			if (Accept (data) != mwOk)
				return mwError;
			continue;
		}

		// if some data was rejected, make a note of it
		if (flags & mwRejected)
		{
			mwData data;
			Recall (data);
			if (Reject (data) != mwOk)
				return mwError;
			continue;
		}

		// stop if the task coordinator failed badly
		if (flags & mwFailed)
			return mwError;

		// stop if the tasks are completed and there is no more data
		if ((flags & mwCompleted) && !data_waiting)
			return mwOk;

		// if data is needed, send it
		if ((flags & mwHungry) && data_waiting)
			if (Send (current) == mwOk)
				data_waiting = false;
	}
} /* mwCoordinator::Coordinate */


} // namespace multiwork
} // namespace chomp

#endif // _MWCOORD_H_

/// @}

