/// @addtogroup multiwork
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file subdiv.h
///
/// This file contains the definition of the MultiWork subdivision framework.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in November 2005. Last revision: March 12, 2006.


#ifndef _MWSUBDIV_H_
#define _MWSUBDIV_H_

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <exception>
#include "chomp/chomp.h"
#include "chomp/multiwork/mw.h"


// --------------------------------------------------
// -------------- the syntax of files ---------------
// --------------------------------------------------

// The syntax of lines in the text data file:
// + n intervals. - data no. n sent
// * n result level cube. - cube no. n received
// # level cube. - cube subdivided

// The syntax of lines in the log file of a coordinator:
// + n. - data no. n prepared
// * n result. - cube no. n received by the coordinator
// ! Message. - data rejected by one of the workers (error)

// The syntax of lines in the log file of a worker:
// - n. - data no. n received by the worker
// = n result. - data no. n sent back from the worker
// ! Message. - data rejected by the worker (error)

namespace chomp {
namespace multiwork {

using namespace chomp::homology;


// --------------------------------------------------
// ------------------- subdivide --------------------
// --------------------------------------------------

/// Subdivides a set of full cubes into twice smaller cubes
/// in each direction.
template <class typeCubes>
void subdivide (const typeCubes &X, typeCubes &result)
{
	typedef typename typeCubes::value_type tCube;
	typedef typename typeCubes::value_type::CoordType tCoord;

	if (X. empty ())
		return;
	int dim = X [0]. dim ();

	tCoord cmin [tCube::MaxDim];
	tCoord cmax [tCube::MaxDim];
	for (int i = 0; i < X. size (); ++ i)
	{
		X [i]. coord (cmin);
		for (int j = 0; j < dim; ++ j)
		{
			cmin [j] *= 2;
			cmax [j] = cmin [j] + 2;
		}
		tRectangle<tCoord> r (cmin, cmax, dim);
		const tCoord *c;
		while ((c = r. get ()) != 0)
			result. add (tCube (c, dim));
	}
	return;
} /* subdivide */

/// Subdivides one full cube into a set of twice smaller cubes
/// in each direction.
template <class typeCube, class typeCubes>
void subdivide (const typeCube &q, typeCubes &result)
{
	typedef typename typeCubes::value_type::CoordType tCoord;

	int dim = q. dim ();

	tCoord cmin [typeCube::MaxDim];
	tCoord cmax [typeCube::MaxDim];
	q. coord (cmin);
	for (int j = 0; j < dim; ++ j)
	{
		cmin [j] *= 2;
		cmax [j] = cmin [j] + 2;
	}
	tRectangle<tCoord> r (cmin, cmax, dim);
	const tCoord *c;
	while ((c = r. get ()) != 0)
		result. add (typeCube (c, dim));
	return;
} /* subdivide */


// --------------------------------------------------
// ---------------- small utilities -----------------
// --------------------------------------------------

/// Creates a full cube of the given dimension with the minimal corner
/// coordinates all zero.
template <class typeCube>
typeCube mwZeroCube (int dim)
{
	typename typeCube::CoordType c [typeCube::MaxDim];
	for (int i = 0; i < dim; ++ i)
		c [i] = 0;
	return typeCube (c, dim);
} /* mwZeroCube */

/// Returns the interval corresponding to the given fraction of an interval.
/// If called with part = 0, ..., full - 1, the returned intervals will cover
/// the provided interval.
template <class coord>
void ifraction (double left, double right, coord part, coord full,
	double &newleft, double &newright)
{
	// uncomment if using CAPD to switch the rounding mode to the nearest
//	round_nearest ();
	if (part)
		newleft = left + (right - left) * part / full;
	else
		newleft = left;
	if (part < full - 1)
		newright = left + (right - left) * (part + 1) / full;
	else
		newright = right;
	return;
} /* ifraction */


// --------------------------------------------------
// ---------------- the worker class -----------------
// --------------------------------------------------

/// The type of a function which computes a value of interest
/// for the given product of intervals. The subdivision level
/// is provided in case the computations on higher levels need
/// to be done in a more thorough way.
typedef int (*fcompute) (const double *left, const double *right,
	int dim, int level);

/// This class defines a worker for the multi-work subdivision framework.
template <int dim, class coord>
class mwSubWorker: public mwWorker
{
public:
	/// The constructor.
	mwSubWorker (int _controlNumber, fcompute _compute);

private:
	/// A function for the initialization of a worker.
	int Initialize (mwData &data);

	/// A function for processing the data by a worker.
	int Process (mwData &data);

	/// A private control number to indentify the coordinator correctly.
	int controlNumber;

	/// The address of a function to use for the computations.
	fcompute compute;

}; /* class mwSubWorker */

// --------------------------------------------------

template <int dim, class coord>
inline mwSubWorker<dim,coord>::mwSubWorker (int _controlNumber,
	fcompute _compute):
	controlNumber (_controlNumber), compute (_compute)
{
	return;
} /* mwSubWorker<dim,coord>::mwSubWorker */

// --------------------------------------------------

template <int dim, class coord>
int mwSubWorker<dim,coord>::Initialize (mwData &data)
{
	// ignore the received data and say it was Ok
	return mwOk;
} /* mwSubWorker<dim,coord>::Initialize */

template <int dim, class coord>
int mwSubWorker<dim,coord>::Process (mwData &data)
{
	// decode the number of the data pack to process
	int current = 0;
	data >> current;

	// decode the subdivision level
	int level = 0;
	data >> level;

	// say which data pack is being processed
	sout << "- " << current << ".\n";

	// the intervals
	double left [dim], right [dim];
	for (int i = 0; i < dim; ++ i)
	{
		data >> left [i];
		data >> right [i];
	}

	// the ending control number
	int ctrl = 0;
	data >> ctrl;
	if (ctrl != controlNumber)
	{
		data. Reset ();
		sout << "! Wrong control number (" << ctrl <<
			"). Rejecting the data.\n";
		return mwReject;
	}

	// rune one set of computations
	int result = compute (left, right, dim, level);

	// if the result is wrong then reject this piece of data
	if (result < 0)
	{
		sout << "! Data no. " << current <<
			" produced a negative result. Rejecting it.\n";
		return mwReject;
	}

	// say what has been computed
	sout << "= " << current << " " << result << ".\n";

	// send back the data containing the result of the processing
	data. Reset ();
	data << current;
	data << result;
	return mwOk;
} /* mwSubWorker<dim,coord>::Process */


// --------------------------------------------------
// ---------------- data pack class -----------------
// --------------------------------------------------

/// This is a helper class which defines a single data pack
/// used in the communication between coordinator and workers
/// in the multi-work subdivision framework.
template <int dim, class coord>
class mwSubDataPack
{
public:
	/// The default constructor.
	mwSubDataPack () {}
	
	/// A nice constructor.
	mwSubDataPack (int _num, const tCubeFix<dim,coord> &_q, int _level):
		num (_num), q (_q), level (_level) {}

	// the default copy constructor and operator = are good

	// no destructor is necessary

	/// The number of the data pack.
	int num;

	/// The cube sent to a worker.
	tCubeFix<dim,coord> q;

	/// The subdivision level of the cube.
	int level;

}; /* struct mwSubDataPack */


// --------------------------------------------------
// ------------- the coordinator class --------------
// --------------------------------------------------

/// This class defines a coordinator for the multi-work subdivision
/// framework.
template <int dim, class coord>
class mwSubCoordinator: public mwCoordinator
{
	/// The maximal feasible subdivision level.
	static const int ppMaxSub = 8 * sizeof (coord) - 2;

	/// The maximal feasible subdivision level.
	static const int ppMaxSubdiv;

	/// The type of a cube of parameters.
	typedef tCubeFix<dim,coord> ppCube;

	/// The maximal dimension of cubes.
	static const int ppMaxDim = ppCube::MaxDim;

	/// The type of a set of cubes in the parameter space.
	typedef hashedset<ppCube> ppCubes;

	/// The kind of rectangle used to iterate ranges of cubes.
	typedef tRectangle<coord> ppRect;

public:
	/// The constructor of an uninitialized coordinator.
	mwSubCoordinator ();

	/// The complete constructor of a coordinator object.
	mwSubCoordinator (const char *filename, int _inilevel, int _maxlevel,
		bool _flushfile, int _controlNumber,
		const double *_paramLeft, const double *_paramRight);

	/// The destructor.
	~mwSubCoordinator ();

private:
	/// A function for preparing data by a coordinator.
	int Prepare (mwData &data);

	/// A function for accepting results by a coordinator.
	int Accept (mwData &data);

	/// A function for taking rejected data by a coordinator.
	int Reject (mwData &data);

	/// Subdivides a given cube into the set of unknown cubes.
	/// Does not remove the cube from its original set, but adds it
	/// to the right 'subdivided' set. Updates 'reslevels'.
	int Subdivide (int level, const ppCube &q);

	/// The current number of data pack to prepare by a coordinator.
	int current;

	/// The set S: subdivided cubes at each subdivision level.
	ppCubes subdivided [ppMaxSub];

	/// The set U: cubes for which the result is yet totally unknown.
	ppCubes unknown [ppMaxSub];

	/// The set Z: cubes which yield the zero result.
	ppCubes zerocubes [ppMaxSub];

	/// The set N: cubes which yield a nonzero result.
	ppCubes nonzerocubes [ppMaxSub];

	/// The cubes which have been sent to workers for processing.
	ppCubes sentcubes [ppMaxSub];
	
	/// The number of levels in the results tables.
	int reslevels;
	
	/// The level of subdivisions in the parameter space at which
	/// the actual computations begin.
	int inilevel;

	/// The maximal allowed level of subdivision in the parameter space.
	int maxlevel;

	/// A file to append the results to.
	std::ofstream f;

	/// Should the results file be flushed after every write?
	bool flushfile;

	/// A procedure for reading the previously computed results.
	int readresults (istream &in);

	/// The left (minimal) vertices of the parameter cube.
	double paramLeft [dim];

	/// The right (maximal) vertices of the parameter cube.
	double paramRight [dim];

	/// The data packs sent for processing.
	multitable<mwSubDataPack<dim,coord> > sent;

	/// The number of data packs sent for processing.
	int sentlen;

	/// A private control number to indentify the workers correctly.
	int controlNumber;

}; /* class mwSubCoordinator */

// --------------------------------------------------

template <int dim, class coord>
const int mwSubCoordinator<dim,coord>::ppMaxSubdiv = 8 * sizeof (coord) - 2;

// --------------------------------------------------

template <int dim, class coord>
inline int mwSubCoordinator<dim,coord>::Subdivide (int level, const ppCube &q)
{
	// make sure the subdivision level will not exceed the maximum
	if (level >= ppMaxSubdiv - 1)
		return 0;
	// update the number of resolution levels used in the sets of cubes
	if (reslevels < level + 2)
		reslevels = level + 2;

	// add the cube to the set of subdivided ones
	subdivided [level]. add (q);

	// subdivide the cube
	subdivide (q, unknown [level + 1]);
	return 0;
} /* mwSubCoordinator<dim,coord>::Subdivide */

// --------------------------------------------------

template <int dim, class coord>
inline mwSubCoordinator<dim,coord>::mwSubCoordinator ()
{
	return;
} /* mwSubCoordinator<dim,coord>::mwSubCoordinator */

template <int dim, class coord>
mwSubCoordinator<dim,coord>::mwSubCoordinator (const char *filename,
	int _inilevel, int _maxlevel, bool _flushfile, int _controlNumber,
	const double *_paramLeft, const double *_paramRight)
{
	// display a warning if no file name is provided
	if (!filename || !*filename)
		sout << "WARNING: No results file name given. "
			"Computation results will not be saved.\n";

	// initialize the subdivision levels
	reslevels = 0;
	inilevel = _inilevel;
	maxlevel = _maxlevel;
	if (maxlevel > ppMaxSubdiv)
	{
		sout << "WARNING: Max subdivision level decreased from " <<
			maxlevel << " to " << ppMaxSubdiv << ".\n";
		maxlevel = ppMaxSubdiv;
	}

	// read the previously computed results of the computations
	if (filename && *filename)
	{
		std::ifstream in (filename);
		if (!in)
			sout << "Note: Could not open the results file. "
			"Any previous results will be ignored.\n";
		else
		{
			readresults (in);
			in. close ();
		}
	}

	// open the results file for appending
	if (filename && *filename)
		f. open (filename, std::ios::out | std::ios::app);
	f << "; Started on " << currenttime ();

	// set up the initial cube in the parameter space if necessary
	if (subdivided [0]. empty () && unknown [0]. empty ())
		unknown [0]. add (mwZeroCube<ppCube> (dim));
	if (reslevels < 1)
		reslevels = 1;

	// set the current number of computations to zero
	current = 0;

	// set the number of sent data to zero
	sentlen = 0;

	// remember the file flushing switch
	flushfile = _flushfile;
	if (flushfile)
		f << std::flush;

	// remember the control number
	controlNumber = _controlNumber;

	// copy the left and right ends of the intervals
	for (int i = 0; i < dim; ++ i)
	{
		paramLeft [i] = _paramLeft [i];
		paramRight [i] = _paramRight [i];
	}
	return;
} /* mwSubCoordinator<dim,coord>::mwSubCoordinator */

template <int dim, class coord>
inline mwSubCoordinator<dim,coord>::~mwSubCoordinator ()
{
	f << "; Finished on " << currenttime ();
	f. close ();
	return;
} /* mwSubCoordinator<dim,coord>::~mwSubCoordinator */

// --------------------------------------------------

/// Finds a cube at the lowest level (>= min, < max) of the given set of
/// cubes that was not sent, yet. Modifies the 'level' and 'q' variables.
/// Returns true if found, false if not found.
template <class typeCubes, class typeCube>
inline bool findmincube (const typeCubes tab [], const typeCubes sent [],
	int minlevel, int maxlevel, int &level, typeCube &q)
{
	for (level = minlevel; level < maxlevel; ++ level)
	{
		int count = tab [level]. size ();
		for (int i = count - 1; i >= 0; -- i)
		{
			q = tab [level] [i];
			if (!sent [level]. check (q))
				return true;
		}
	}
	return false;
} /* findmincube */

/// Finds a cube at the highest level (>= min, < max) of the given set of
/// cubes that was not sent, yet. Modifies the 'level' and 'q' variables.
/// Returns true if found, false if not found.
template <class typeCubes, class typeCube>
inline bool findmaxcube (const typeCubes tab [], const typeCubes sent [],
	int minlevel, int maxlevel, int &level, typeCube &q)
{
	for (level = maxlevel - 1; level >= minlevel; -- level)
	{
		int count = tab [level]. size ();
		for (int i = count - 1; i >= 0; -- i)
		{
			q = tab [level] [i];
			if (!sent [level]. check (q))
				return true;
		}
	}
	return false;
} /* findmaxcube */

template <int dim, class coord>
int mwSubCoordinator<dim,coord>::Prepare (mwData &data)
{
	// determine the data pack to send
	int level;
	ppCube q;
	while (1)
	{
		// find an unknown cube at the initial subdivision level
		if (findmincube (unknown, sentcubes, inilevel, inilevel + 1,
			level, q))
			break;
		
		// try and subdivide one of big cubes to obtain smaller ones
		// (either an unknown cube, or a zero cube if any)
		if (findmaxcube (unknown, sentcubes, 0, inilevel,
			level, q) || findmaxcube (zerocubes, sentcubes,
			0, inilevel, level, q))
		{
			f << "# " << level << " " << q << ".\n";
			if (flushfile)
				f << flush;
			Subdivide (level, q);
			unknown [level]. remove (q);
			zerocubes [level]. remove (q);
			continue;
		}

		// find an unknown cube at a decent subdivision level
		if (findmincube (unknown, sentcubes, inilevel, reslevels,
			level, q))
			break;

		// if no unknown cube could be found, then find a zero cube
		// and subdivide it
		if (findmincube (zerocubes, sentcubes,
			inilevel, maxlevel, level, q))
		{
			f << "# " << level << " " << q << ".\n";
			if (flushfile)
				f << flush;
			Subdivide (level, q);
			zerocubes [level]. remove (q);

			continue;
		}

		// if no cube for processing can be found, then interrupt
		return mwNoData;
	}

	// add the cube to the set of data that was sent
	sentcubes [level]. add (q);
	sent [sentlen ++] = mwSubDataPack<dim,coord> (current, q, level);

	// save the information on this data to the results file
	f << "+ " << current;

	// prepare the intervals of parameters corresponding to the cube
	coord c [ppMaxDim];
	q. coord (c);
	int full = 1 << level;
	double left [ppMaxDim], right [ppMaxDim];
	for (int i = 0; i < dim; ++ i)
	{
		ifraction (paramLeft [i], paramRight [i], c [i], full,
			left [i], right [i]);
		f << " [" << left [i] << "," << right [i] << "]";
	}
	f << ".\n";

	// flush the data file if requested to
	if (flushfile)
		f << flush;

	// encode the current number to be sent
	data << current;
	
	// encode the subdivision level
	data << level;

	// encode the intervals (left and right ends)
	for (int i = 0; i < dim; ++ i)
		data << left [i] << right [i];

	// append the control number for extra verification
	data << controlNumber;

	// show a message on what has been done
	sout << "+ " << current << ".\n";

	++ current;
	return mwOk;
} /* mwSubCoordinator<dim,coord>::Prepare */

template <int dim, class coord>
int mwSubCoordinator<dim,coord>::Accept (mwData &data)
{
	// decode the number
	int n = 0;
	data >> n;
	int result = 0;
	data >> result;

	// say what has been received
	sout << "* " << n << " " << result << ".\n";

	// find this data among the data that was sent
	int pos = 0;
	while ((pos < sentlen) && (sent [pos]. num != n))
		++ pos;
	if (pos >= sentlen)
		throw "Wrong returned data number.";

	// save the data to the output file
	f << "* " << n << " " << result <<
		" " << sent [pos]. level << " " << sent [pos]. q << ".\n";

	// flush the data file if requested to
	if (flushfile)
		f << flush;
	
	// process the received data as necessary
	int level = sent [pos]. level;
	const ppCube &q = sent [pos]. q;
	sentcubes [level]. remove (q);
	if (result)
	{
		unknown [level]. remove (q);
		nonzerocubes [level]. add (q);
	}
	else
	{
		unknown [level]. remove (q);
		zerocubes [level]. add (q);
	}

	// remove the data from the table containing the data that was sent
	-- sentlen;
	if (pos != sentlen)
		sent [pos] = sent [sentlen];

	return mwOk;
} /* mwSubCoordinator<dim,coord>::Accept */

template <int dim, class coord>
int mwSubCoordinator<dim,coord>::Reject (mwData &data)
{
	// decode the number
	int n = 0;
	data >> n;

	// say that data has been rejected
	sout << "!!! WARNING !!! Data no. " << n << " rejected.\n";

	return mwOk;
} /* mwSubCoordinator<dim,coord>::Reject */

template <int dim, class coord>
int mwSubCoordinator<dim,coord>::readresults (istream &in)
{
	ignorecomments (in);
	while (!in. eof ())
	{
		// read the first char and interprete the line
		int ch = in. get ();

		// ignore the line if it contains no useful information
		if ((ch != '*') && (ch != '#'))
		{
			ignoreline (in);
			ignorecomments (in);
			continue;
		}
	
		// read the data number if relevant
		if (ch == '*')
		{
			int n = -1;
			in >> n;
			if (n < 0)
				break;
		}

		// read the result if any
		int result = -1;
		if (ch == '*')
		{
			in >> result;
			if (result < 0)
				break;
		}

		// read the subdivision level of the cube
		int level = -1;
		in >> level;
		if (level < 0)
			break;

		// read the cube itself
		ppCube q;
		in >> q;

		// make sure that there is the dot which ends the line
		if (in. peek () != '.')
			break;

		// ignore the remaining part of the line
		ignoreline (in);
		ignorecomments (in);

		// process the data as necessary
		if (ch == '*')
		{
			if (result)
			{
				unknown [level]. remove (q);
				nonzerocubes [level]. add (q);
			}
			else
			{
				unknown [level]. remove (q);
				zerocubes [level]. add (q);
			}
		}
		else if (ch == '#')
		{
			Subdivide (level, q);
		}

		// update the number of subdivision levels
		if (reslevels <= level)
			reslevels = level + 1;
	}

	// remove subdivided cubes from all the other sets
	for (int level = 0; level < reslevels; ++ level)
	{
		unknown [level]. remove (subdivided [level]);
		zerocubes [level]. remove (subdivided [level]);
		nonzerocubes [level]. remove (subdivided [level]);
	}

	return 0;
} /* mwSubCoordinator<dim,coord>::readresults */


// --------------------------------------------------
// ---------------------- MAIN ----------------------
// --------------------------------------------------

/// The main procedure for running the computations in the multiwork
/// subdivision framework.
/// Returns: 0 = Ok, -1 = Error, 1 = Help displayed, 2 = Wrong arguments.
template <int dim, class coord>
int mwSubdivMain (int argc, char *argv [],
	const char *title, const char *helpinfo,
	int defaultPortNumber, int controlNumber,
	const double *paramLeft, const double *paramRight,
	fcompute compute)
{
	const char *arginfo = "\
Command line arguments (at least '-w' or '-m N' must be specified):\n\
-w - run as a worker (by default the program runs as a coordinator),\n\
-c - run as a coordinator only (don't process any data locally),\n\
-k - keep workers waiting after the computations have been completed,\n\
-p N - set the port number to N (use 0 to reject incoming connections),\n\
computer:port - use this connection at start-up (can be repeated),\n\
-s FILE - save the workers' list to this file,\n\
-r FILE - retrieve the workers' list from this file,\n\
-f filename - results file (new results will be appended); coord only!\n\
--flush - flush the results file very frequently (slows down the program!)\n\
-i N - set the initial parameter cube subdivision level (default: 0),\n\
-m N - set the maximal subdivision level,\n\
-q - quit all the workers who are waiting,\n\
--quiet - do not display any messages on the standard output,\n\
--log filename - save the console output to the given file,\n\
--help - show this brief help information and exit.\n\
For more information ask the author at http://www.PawelPilarczyk.com/.";

	// prepare user-configurable data
	char *retrieveworkers = 0;
	char *saveworkers = 0;
	const int maxaddr = 1024;
	char *addr [maxaddr];
	int naddr = 0;
	int port = defaultPortNumber;
	bool localwork = true;
	bool worker = false;
	bool keepworkers = false;
	bool quitworkers = false;
	char *filename = 0;
	int inilevel = 0;
	int maxlevel = 0;
	bool flushfile = false;

	// interprete the command-line arguments
	arguments a;
	arg (a, 0, addr, naddr, maxaddr);
	arg (a, "r", retrieveworkers, "mwsubdiv.txt");
	arg (a, "s", saveworkers, "mwsubdiv.txt");
	arg (a, "p", port);
	arg (a, "f", filename);
	arg (a, "i", inilevel);
	arg (a, "m", maxlevel);
	argswitch (a, "c", localwork, false);
	argswitch (a, "w", worker, true);
	argswitch (a, "k", keepworkers, true);
	argswitch (a, "q", quitworkers, true);
	argswitch (a, "-flush", flushfile, true);
	arghelp (a);

	argstreamprepare (a);
	int argresult = a. analyze (argc, argv);
	argstreamset ();

	// show the program's main title
	if (argresult >= 0)
		sout << title << '\n';

	// if something was incorrect, show an additional message and exit
	if (argresult < 0)
	{
		sout << "Call with '--help' for help.\n";
		return 2;
	}

	// if no data packs are to be processed, don't run the program
	if (!maxlevel && !worker && !quitworkers)
		argresult = 1;
	
	// if help requested, show help information
	if (argresult > 0)
	{
		sout << helpinfo << '\n' << arginfo << '\n';
		return 1;
	}

	// try running the main function and catch an error message if thrown
	try
	{
		// set an appropriate program time message
		program_time = "Aborted after:";
		program_time = 1;

		// quit all the workers from the list if requested to
		if (quitworkers)
		{
			mwSubCoordinator<dim,coord> c;
			if (retrieveworkers)
				c. Load (retrieveworkers);
			for (int i = 0; i < naddr; ++ i)
				c. Add (addr [i]);
			sout << "Quitting workers... ";
			c. QuitWorkers ();
			sout << "Done.\n";
		}

		// run as a worker if requested to
		else if (worker)
		{
			mwSubWorker<dim,coord> w (controlNumber, compute);
			for (int i = 0; i < naddr; ++ i)
				w. Add (addr [i]);
			w. Port (port);
			sout << "Running as a worker...\n";
			int result = w. Work ();
			if (result == mwOk)
				sout << "Work completed successfully.\n";
			else
				sout << "Could not work - probably "
					"an error occurred.\n";
		}

		// run as a coordinator otherwise
		else
		{
			mwSubWorker<dim,coord> w (controlNumber, compute);
			mwSubCoordinator<dim,coord> c (filename, inilevel,
				maxlevel, flushfile, controlNumber,
				paramLeft, paramRight);
			c. KeepWorkers (keepworkers);
			for (int i = 0; i < naddr; ++ i)
				c. Add (addr [i]);
			if (retrieveworkers)
				c. Load (retrieveworkers);
			c. Port (port);
			sout << "Running as a coordinator...\n";
			int result = c. Coordinate (localwork ? &w : 0);
			if (result == mwOk)
				sout << "The task completed successfully.\n";
			else
				sout << "Could not coordinate - probably "
					"an error occurred.\n";
			if (saveworkers)
				c. Save (saveworkers);
		}

		// set an appropriate program time message
		program_time = "Total time used:";

		// finalize
		return 0;
	}
	catch (const char *msg)
	{
		sout << "ERROR: " << msg << '\n';
		return -1;
	}
	catch (const std::exception &e)
	{
		sout << "ERROR: " << e. what () << '\n';
		return -1;
	}
	catch (...)
	{
		sout << "ABORT: An unknown error occurred.\n";
		return -1;
	}
} /* mwSubdivMain */


} // namespace multiwork
} // namespace chomp

#endif // _MWSUBDIV_H_

/// @}

