/// @addtogroup multiwork
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file task.h
///
/// This file contains the definition of the MultiWork task class.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on August 11, 2004. Last revision: January 29, 2006.


#ifndef _MWTASK_H_
#define _MWTASK_H_

#include <fstream> // for the log file
#include <cstdlib> // for "atoi"
#include <ctime> // for log start/stop time
#include <cstring> // for "strlen" and "strcpy"

#include "chomp/multiwork/data.h"


// --------------------------------------------------
// -------------------- settings --------------------
// --------------------------------------------------

// are we in a network environment (otherwise this is a single process)
#ifndef mwNETWORK
	#define mwNETWORK 1
#endif // mwNETWORK

/// Unexpected errors occur while using wxWin sockets from within the Krak
/// library from the CAPD package, so I turn this feature off. This may be
/// due to the fact that the main program runs there as a secondary thread.
/// This should be fixed in the future to make the program run in WXGTK.
#define mwWXWIN 0

// should the wxWindows sockets be used?
#ifndef mwWXWIN
	#if (defined (__WXMSW__) || defined (__WXGTK__) || \
		defined (__WXMOTIF__) || defined (__WXX11__) || \
		defined (__WXMAC__) || defined (__WXUNIVERSAL__))
		#define mwWXWIN 1
	#else
		#define mwWXWIN 0
	#endif
#endif // mwWXWIN

// the default port number
#ifndef mwPORT
	#define mwPORT 5885
#endif // mwPORT
	
// the default control number used for identification
#ifndef mwCTRLNUM
	#define mwCTRLNUM 16780
#endif // mwCTRLNUM

// the default time-out in seconds (8 hours)
#ifndef mwTIMEOUT
	#define mwTIMEOUT 28800
#endif // mwTIMEOUT

// the maximal number of allowed workers
#ifndef mwMAXWORK
	#define mwMAXWORK 1024
#endif // mwMAXWORK

// the maximal number of data elements to be stored in the task object
#ifndef mwMAXDATA
	#define mwMAXDATA (mwMAXWORK + 16)
#endif // mwMAXDATA

// the maximal number of the initialization data pieces that can be sent
#ifndef mwMAXINIT
	#define mwMAXINIT 256
#endif // mwMAXINIT

// is the worker class included in the compilation? [unused for the moment]
#ifndef mwWORKER
	#define mwWORKER 1
#endif // mwWORKER

// is the coordinator class included in the compilation? [currently unused]
#ifndef mwCOORDINATOR
	#define mwCOORDINATOR 1
#endif // mwCOORDINATOR


// --------------------------------------------------
// ---------------------- bool ----------------------
// --------------------------------------------------

// if the 'bool' data type seems to be unknown to the compiler,
// define your own 'bool', 'true' and 'false' to use in the program
#if !mwBOOL
	#define bool unsigned char
	#define true 1
	#define false 0
#endif // no mwBOOL


namespace chomp {
namespace multiwork {

// --------------------------------------------------
// ------------------- constants --------------------
// --------------------------------------------------

enum mwConstants
{
	// various error codes (must be negative!)
	mwError = -1,	// any kind of error
	mwNoData = -2,	// no data to send to workers
	mwReject = -3,	// reject the data
	mwTimeOut = -4,	// connection timed-out
	mwLost = -5,	// connection lost

	// everything is Ok (should be zero)
	mwOk = 0
}; /* enum mwConstants */

enum mwTaskBits
{
	mwNone = 0x0000,
	mwNoMoreData = 0x0001,
	mwCompleted = mwNoMoreData,
	mwArrived = 0x0002,
	mwRejected = 0x0004,
	mwHungry = 0x0008,
	mwFailed = 0x0010
}; /* enum mwTaskBits */

enum mwLogLevel
{
//	mwNone = 0,
	mwRough = 1,
	mwMedium,
	mwDetailed,
	mwDebug
}; /* enum mwLogLevel */


// --------------------------------------------------
// ---------- network communication flags -----------
// --------------------------------------------------

enum mwIOflags
{
	mwCanRead = 0x01,
	mwCanWrite = 0x02
}; /* enum mwIOflags */

enum mwCode
{
	mwNoMsg = 0x0000,	// no special code info included
	// messages sent to the worker only
	mwInitMsg = 0x0001,	// this is initialization data
	mwStdMsg = 0x0002,	// standard data to be processed
	mwKeepMsg = 0x0004,	// keep running after having disconnected
	mwDontKeepMsg = 0x0008,	// don't keep running after disconnected
	mwByeMsg = 0x0010,	// disconnect, please!
	// messages sent to the coordinator only
	mwRejectedMsg = 0x0001,	// the data was rejected
	mwPortMsg = 0x0002,
	// control code shift and message code mask
	mwShiftMsg = 5,
	mwMaskMsg = 0x001F
}; /* enum mwCode */


// --------------------------------------------------
// --------------------- mwTask ---------------------
// --------------------------------------------------

/// This class defines a generic task object (coordinator or worker)
/// for the multi-work distributed computations framework.
/// It is the common part of a worker task and a coordinator
/// task, and contains some general settings that apply to both.
/// Note that this class is inherited by the worker and the coordinator
/// class in the virtual mode, which allows one to create a class that is
/// both a worker and a coordinator in one (although this is not
/// recommended in general).
class mwTask
{
public:
	// --- constructor/destructor ---
	
	// the default constructor
	mwTask ();
		
	// the destructor
	~mwTask ();

	// --- configuration ---

	// set the port number for the communication or 0 to use none;
	// return the current port number (<0 = only return the number)
	int Port (int number);

	// set the control number used for identification; at most 27 lower
	// bits (or maybe even less) of this number are actually verified
	void ControlNumber (int number);

	// set the time-out interval in seconds; this is used by a worker
	// or a coordinator to wait for a network connection by the other one
	void TimeOut (int seconds);

	// --- log file ---

	// log communication information to a file; replace the filename's
	// extension with the given one (works for Unix and Windows names);
	// return mwOk on success, mwError if cannot open/create the file
	int LogFile (const char *filename, const char *extension = NULL);

	// use another task's log file to log this task's information
	void LogFile (const mwTask &other);

	// set the level of log messages: mwRough, mwMedium, mwDetailed;
	// return the current log level
	int LogLevel (int level = -1);
	
	// append a message to the log file and add '\n' if requested to
	int Log (const char *message = NULL, bool newLine = true);

	// close the log file and add a line with the time information
	void LogClose ();

	// --- computer addresses ---

	// add another address to the list of computers to work with;
	// the addresses can be in the form "computer.domain:port"
	int Add (const char *name, int port = -1);
	
	// load computer addresses from a file; replace the filename's
	// extension with the given one (works for Unix and Windows names);
	// return the number of addresses read or mwError
	int Load (const char *filename, const char *extension = NULL);
	
	// --- quit waiting workers ---

	// quit all the workers whose addresses were added with the 'Add'
	// and 'Load' functions; return mwOk or mwError
	int QuitWorkers ();

	// --- other ---

	// replace the filename's extension (if any) with the given one;
	// the returned name is allocated with 'new' and must be deleted []
	static char *SetExtension (const char *filename, const char *ext);

	// remove path from the file name and return a pointer to the
	// filename without extensin within the given file name
	static const char *WithoutPath (const char *filename);

protected:
	// --- configuration data ---
	
	// the network communication port number
	int portnum;

	// the control number that is used to recognize a compatible worker
	// or a compatible coordinator; note: the worker sends its negation
	unsigned int ctrlnum;
	unsigned int negctrlnum;

	// the time-out in seconds
	int timeout;

	// --- log file ---

	// log file stream and the level of details
	std::ofstream *logFile;
	int logLevel;

	// is this log file pointer borrowed from another task?
	bool logBorrowed;

	// append a message to the log file and add '\n' if requested to;
	// this tricky form is used to write to '*logFile' directly and
	// to put it in one instruction, eg.: Log (*logFile << "x = " << 13);
	int Log (const void *, bool newLine = true);

	// --- computer addresses ---

	// a list of workers or coordinators to connect to at start-up
	mwData computers;
	
private:
	// --- no copying constructor or assignment operator ---
	
	// the copying constructor is forbidden
	mwTask (const mwTask &) {}
		
	// the assignment operator is forbidden
	mwTask &operator = (const mwTask &) {return *this;}
		
}; /* class mwTask */

// --------------------------------------------------

inline mwTask::mwTask ():
	portnum (mwPORT),
	ctrlnum (mwCTRLNUM),
	negctrlnum (~mwCTRLNUM),
	timeout (mwTIMEOUT),
	logFile (NULL),
	logLevel (mwMedium),
	logBorrowed (false)
{
	return;
} /* mwTask::mwTask */

// --------------------------------------------------

inline char *mwTask::SetExtension (const char *filename, const char *ext)
{
	// if there is no file name, return NULL
	if (!filename)
		return NULL;
	
	// determine the lengths of both parts of the file name
	int namelen = strlen (filename);
	int extlen = ext ? strlen (ext) : 0;

	// allocate memory for the new name and copy the first part
	char *name = new char [namelen + extlen + 2];
	strcpy (name, filename);

	// if no extension change is necessary, the result is obvious
	if (!ext)
		return name;

	// determine the position in the file name to start at
	int pos = namelen ? (namelen - 1) : 0;

	// skip all the periods at the very end of the file name
	while (pos && (name [pos] == '.'))
		-- pos;

	// check if the name has an extension to remove
	while (pos && (name [pos] != '.') && (name [pos] != ':') &&
		(name [pos] != '/') && (name [pos] != '\\'))
		-- pos;

	// truncate the name to remove the unwanted extension
	if (pos && (name [pos] == '.'))
		namelen = pos;

	// add the period before the extension if necessary
	if (*ext && (*ext != '.'))
		name [namelen ++] = '.';

	// append the extension to the file name
	strcpy (name + namelen, ext);
	
	// return the new name
	return name;
} /* mwTask::SetExtension */

inline const char *mwTask::WithoutPath (const char *filename)
{
	// if there is no file name, return NULL
	if (!filename)
		return NULL;
	
	// determine the lengths of both parts of the file name
	int pos = strlen (filename);
	while ((pos > 0) && (filename [pos - 1] != '\\') &&
		(filename [pos - 1] != '/') && (filename [pos - 1] != ':'))
		-- pos;
	return filename + pos;
} /* mwTask::WithoutPath */

// --------------------------------------------------

inline void mwTask::LogClose ()
{
	if (!logFile)
		return;
	if (logBorrowed)
	{
		logFile = NULL;
		logBorrowed = false;
		return;
	}
	std::time_t stop_time;
	std::time (&stop_time);
	*logFile << "\nMultiWork log file closed on " <<
		std::asctime (std::localtime (&stop_time)) << "\n"
	"-----------------------------------------------------\n\n";
	delete logFile;
	logFile = NULL;
	return;
} /* mwTask::LogClose */

inline int mwTask::LogFile (const char *filename, const char *extension)
{
	// close the current log if in use
	LogClose ();

	// if no file name supplied, return now
	if (!filename)
		return mwOk;

	// create a file stream variable
	logFile = new std::ofstream;
	if (!logFile)
		return mwError;

	// open the log file for appending
	char *name = SetExtension (filename, extension);
	logFile -> open (name, std::ios::app);
	delete [] name;

	// write the current time to the log
	std::time_t start_time;
	std::time (&start_time);
	*logFile << "MultiWork log file opened on " <<
		std::asctime (std::localtime (&start_time)) << "\n" <<
		std::flush;

	// if unable to open the file or an error occurred, return mwError
	if (!*logFile)
	{
		delete logFile;
		return mwError;
	}

	return mwOk;
} /* mwTask::LogFile */

inline void mwTask::LogFile (const mwTask &other)
{
	// close the current log if in use
	LogClose ();

	// if there is no other log file, do nothing
	if (!other. logFile)
		return;

	// borrow the log file pointer
	logBorrowed = true;
	logFile = other. logFile;
	logLevel = other. logLevel;

	return;
} /* mwTask::LogFile */

inline int mwTask::LogLevel (int level)
{
	if (level >= 0)
		logLevel = level;
	return logLevel;
} /* mwTask::LogLevel */
	
inline int mwTask::Log (const char *message, bool newLine)
{
	if (!logFile)
		return mwError;
	if (message && *message)
		*logFile << message;
	if (newLine)
		*logFile << '\n';
	if (newLine || (logLevel >= mwMedium))
		logFile -> flush ();
	if (!*logFile)
	{
		if (!logBorrowed)
			delete logFile;
		logFile = NULL;
		logBorrowed = false;
		return mwError;
	}
	return mwOk;
} /* mwTask::Log */

inline int mwTask::Log (const void *, bool newLine)
{
	return Log ((const char *) NULL, newLine);
} /* mwTask::Log */

// --------------------------------------------------

inline mwTask::~mwTask ()
{
	// close the log file if it was in use
	LogClose ();

	return;
} /* mwTask::~mwTask */

// --------------------------------------------------

inline int mwTask::Port (int number)
{
	if (number >= 0)
		portnum = (short int) number;
	return portnum;
} /* mwTask::Port */

inline void mwTask::ControlNumber (int number)
{
	ctrlnum = (unsigned int) number;
	negctrlnum = ~ctrlnum;
	return;
} /* mwTask::ControlNumber */

inline void mwTask::TimeOut (int seconds)
{
	timeout = seconds;
	return;
} /* mwTask::TimeOut */

// --------------------------------------------------

inline int mwTask::Add (const char *name, int port)
{
	// if the name is empty, ignore it
	if (!name || !*name)
		return mwError;

	// determine whether the name contains a colon and port number
	int pos = 1;
	while (name [pos])
		pos ++;
	pos --;
	while (pos && (name [pos] != ':') &&
		(name [pos] >= '0') && (name [pos] <= '9'))
		pos --;

	// if the name contains colon and some digits after the colon...
	if (pos && (name [pos] == ':') && name [pos + 1])
	{
		// append the computer name and the chosen port number
		computers. Append (name, pos);
		computers. Append ('\0');
		computers. Append (std::atoi (name + pos + 1));
	}
	else
	{
		// if the port number is not reasonable, use the default one
		if (port <= 0)
			port = portnum;
		computers << name << port;
	}
	return mwOk;
} /* mwTask::Add */

inline int mwTask::Load (const char *filename, const char *extension)
{
	char *name = SetExtension (filename, extension);
	std::ifstream f (name);
	delete [] name;
	if (!f)
		return mwError;

	char buf [512];
	int counter = 0;
	while (1)
	{
		*buf = '\0';
		f. getline (buf, 512, '\n');
		if ((*buf == ';') || (*buf == '#') || (*buf == '/'))
			continue;
		if (*buf)
		{
			Add (buf);
			++ counter;
		}
		if (!f)
			return counter;
	}
} /* mwTask::Load */


} // namespace multiwork
} // namespace chomp

#endif // _MWTASK_H_

/// @}

