/// @addtogroup multiwork
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file worker.h
///
/// This file contains the definition of the MultiWork worker class.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on August 11, 2004. Last revision: January 13, 2006.


#ifndef _MWWORKER_H_
#define _MWWORKER_H_

#include "chomp/multiwork/task.h"

namespace chomp {
namespace multiwork {

class mwCoordinator;


// --------------------------------------------------
// -------------------- mwWorker --------------------
// --------------------------------------------------

/// This class defines a generic worker task object for the multi-work
/// distributed computations framework.
/// Each worker class defined by the user must inherit from this class.
class mwWorker: public virtual mwTask
{
public:
	// --- constructor/destructor ---
	
	// the default constructor
	mwWorker ();
		
	// the destructor
	virtual ~mwWorker ();

	// --- run the worker ---

	// run the worker on this computer; return mwOk or mwError
	int Work ();
		
	// --- configuration ---

	// keep the worker running after the coordinator has disconnected
	void KeepWorker (bool keep = true);

protected:
	// the address of a local coordinator if working locally,
	// NULL otherwise
	mwCoordinator *localCoord;

private:
	// --- worker's job ---

	// acquire a portion of the initialization data;
	// return mwOk on success, mwError to quit the program;
	// overload this function in the class you derive from mwWorker
	virtual int Initialize (mwData &data);
		
	// process a data portion and replace it with the result;
	// return mwOk on success, mwReject to reject the data,
	// or mwError to quit the program;
	// overload this function in the class you derive from mwWorker
	virtual int Process (mwData &data);

	// --- configuration ---

	// should the worker remain running after coordinator disconnects?
	bool keepWorker;

	// grant access to the protected functions and data to a coordinator
	friend class mwCoordinator;
	
}; /* class mwWorker */

// --------------------------------------------------

inline mwWorker::mwWorker (): localCoord (NULL), keepWorker (false)
{
	return;
} /* mwWorker::mwWorker */

inline mwWorker::~mwWorker ()
{
	return;
} /* mwWorker::~mwWorker */

// --------------------------------------------------

inline void mwWorker::KeepWorker (bool keep)
{
	keepWorker = keep;
	return;
} /* mwWorker::KeepWorkers */

// --------------------------------------------------

inline int mwWorker::Initialize (mwData &)
{
	return mwOk;
} /* mwWorker::Initialize */

inline int mwWorker::Process (mwData &)
{
	return mwOk;
} /* mwWorker::Process */


} // namespace multiwork
} // namespace chomp

#endif // _MWWORKER_H_

/// @}

