/////////////////////////////////////////////////////////////////////////////
///
/// @file homcubes.cpp
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on August 10, 2004. Last revision: September 2, 2004.


// --------------------------------------------------
// ------------------ header files ------------------
// --------------------------------------------------

// the multi-work library header
#include "chomp/multiwork/mw.h"
using namespace chomp::multiwork;

// various system headers
#if defined (__WIN32__)
#include <windows.h> // for "Sleep"
#endif

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <unistd.h> // for "sleep"
#include <cstdlib> // for "atoi" and "rand"


// --------------------------------------------------
// ------------- title and information --------------
// --------------------------------------------------

static const char *title = "\
Multi-Work demo: Numbers. Copyright (C) 2004 by Pawel Pilarczyk.\n\
This is free software. No warranty. Consult 'license.txt' for details.";

static const char *helpinfo = "\
This program is a simple demonstration of the Multi-Work interface.\n\
To see the effects, run one copy of this program on your computer,\n\
and then run sevral other copies with the arguments '-w localhost'.\n\
You can also run the other copies on different machines, and then replace\n\
'localhost' with the name of the computer running the first program.\n\
The first program acts as a coordinator who sends numbers to workers,\n\
and the other programs act as workers who process the numbers and send\n\
back their answers.\n\
Command line arguments:\n\
-w - run as a worker (by default the program runs as a coordinator),\n\
-c - run as a coordinator only (don't process any data locally),\n\
-k - keep workers running and waiting when the computations are done,\n\
-fNUM - set the first number to be used in the test,\n\
-mNUM - set the maximal number to be used in the test,\n\
-oN - set the connection time-out to N seconds,\n\
-pN - set the port number to N (use 0 to reject all incoming connections),\n\
-lFILE - log details to this file,\n\
computer:port - use this connection at start-up (can be repeated),\n\
-sFILE - save the workers' list to this file,\n\
-rFILE - retrieve the workers' list from this file,\n\
-q - quit all the workers who are waiting\n\
-h - show this brief help information and exit.\n\
For more details ask the author at http://www.PawelPilarczyk.com/.";


// --------------------------------------------------
// ----------------- the task class -----------------
// --------------------------------------------------
// Note: This class is both a worker and a coordinator in one.
// Normally, these should be two distinct classes, but this also works.
// Please, consult other examples for that possibility.

class Numbers: public mwWorker, public mwCoordinator
{
public:
	// default constructor
	Numbers ();

	// --- the coordinator's data ---
	
	// the first and maximal numbers to send
	int first, maximal;
		
private:
	// --- the worker's functions and data ---

	// a function for the initialization of a worker
	int Initialize (mwData &x);

	// a function for processing the data as a worker
	int Process (mwData &x);
		
	// the lucky number received by a worker
	int luckynumber;
		
	// --- the coordinator's functions and data ---
	
	// a function for preparing data by a coordinator
	int Prepare (mwData &data);

	// a function for accepting results by a coordinator
	int Accept (mwData &data);

	// a function for taking rejected data by a coordinator
	int Reject (mwData &data);

	// the number to be sent by a coordinator
	int current;

}; /* class Numbers */

// --------------------------------------------------

Numbers::Numbers ()
{
	luckynumber = 7;
	current = -1;
	first = 1;
	maximal = 14;
	return;
} /* Numbers::Numbers */

// --------------------------------------------------

int Numbers::Initialize (mwData &x)
{
	// retrieve the initialization data
	x >> luckynumber;

	// if the data is wrong, return the error code
	if (luckynumber <= 0)
	{
		std::cout << "Wrong lucky number received." << std::endl;
		return mwError;
	}

	// accept the initialization data and return the Ok code
	std::cout << "The lucky number set to " << luckynumber << "." <<
		std::endl;
	return mwOk;
} /* Numbers::Initialize */

int Numbers::Process (mwData &x)
{
	// retrieve the number sent and make it non-negative
	int nr;
	x >> nr;
	if (nr < 0)
		nr = -nr;

	// select an arbitrary sleep-time depending on the data
	int sleeptime = (rand () % (700 + nr)) % 9 + 1;

	// say that the data is being processed
	std::cout << "Processing " << nr << " (" << sleeptime << " sec)... " <<
		std::flush;
		
	// process the data (only sleep a few seconds in this example)
	#if defined (__WIN32__)
	Sleep (1000 * sleeptime);
	#else
	sleep (sleeptime);
	#endif
	
	// if the result should be rejected, return the rejection code
	if (!(nr % luckynumber))
	{
		std::cout << "Divides by " << luckynumber <<
			". Rejecting." << std::endl;
		return mwReject;
	}

	// send back the data containing the result of the processing
	nr = -nr;
	std::cout << "Sending back " << nr << "." << std::endl;
	x. Reset ();
	x << nr;
	return mwOk;
} /* Numbers::Process */

// --------------------------------------------------

int Numbers::Prepare (mwData &data)
{
	// initialize the current number if this is the first run
	if (current < 0)
		current = first;

	// if all the numbers have been sent already, say that
	if (current > maximal)
		return mwNoData;

	// send the current number and increase it
	std::cout << "Sending number " << current << "." << std::endl;
	data << current ++;

	return mwOk;
} /* Numbers::Prepare */

int Numbers::Accept (mwData &data)
{
	// decode the number
	int n = 0;
	data >> n;

	// say what was received
	std::cout << "Received number " << n << "." << std::endl;

	return mwOk;
} /* Numbers::Accept */

int Numbers::Reject (mwData &data)
{
	// decode the number
	int n = 0;
	data >> n;

	// say what was received
	std::cout << "This number was rejected: " << n << "." << std::endl;

	return mwOk;
} /* Numbers::Reject */


// --------------------------------------------------
// ---------------------- main ----------------------
// --------------------------------------------------

int main (int argc, char *argv [])
{
	// show the program's title
	std::cout << title << std::endl;
	
	// create one instance of the task class
	Numbers numbers;
	bool localWork = true;

	// set the time-out to one hour
	numbers. TimeOut (3600);

	// analyze the command-line arguments
	bool worker = false;
	char *savename = NULL;
	bool quit = false;
	for (int i = 1; i < argc; ++ i)
	{
		// if this argument does not begin with '-'...
		if (argv [i] [0] != '-')
		{
			numbers. Add (argv [i]);
			continue;
		}

		// if the argument is just '-'...
		if (!argv [i] [1])
		{
			continue;
		}

		// determine the right option name and its value
		char option = argv [i] [1];
		char *value = argv [i] + 2;
		if ((option != 'w') && (option != 'c') && (option != 'k') &&
			 (option != 'q') && !*value && (i + 1 < argc) &&
			(argv [i + 1] [0] != '-'))
			value = argv [++ i];

		// interprete the option and its value
		switch (option)
		{
		case 'w':
			worker = true;
			break;
		case 'c':
			localWork = false;
			break;
		case 'l':
			numbers. LogFile (value);
			numbers. LogLevel (mwDebug);
			break;
		case 'p':
			numbers. Port (atoi (value));
			break;
		case 'o':
			numbers. TimeOut (atoi (value));
			break;
		case 'k':
			if (*value == '0')
				numbers. KeepWorkers (false);
			else
				numbers. KeepWorkers (true);
			break;
		case 's':
			savename = value;
			break;
		case 'r':
			numbers. Load (value);
			break;
		case 'f':
			numbers. first = atoi (value);
			if (numbers. first < 0)
				numbers. first = 0;
			break;
		case 'm':
			numbers. maximal = atoi (value);
			break;
		case 'q':
			quit = true;
			break;
		case '-':
		case 'h':
		case '?':
			numbers. Log ("Help info displayed.");
			std::cout << helpinfo << std::endl;
			return 0;
		default:
			break;
		}
	}

	// quit all the workers if this is the only thing requested
	if (quit)
	{
		numbers. QuitWorkers ();
	}

	// run as a coordinator otherwise
	else if (!worker)
	{
		// send some initialization data to all the workers
		mwData init;
		init << 5;
		numbers. Init (init);

		std::cout << "Running as a coordinator..." << std::endl;
		int result = numbers. Coordinate (localWork ? &numbers :
			NULL);
		if (result == mwOk)
			std::cout << "Done." << std::endl;
		else
			std::cout << "The computations could not be "
				"completed." << std::endl;

		// save the workers and their port numbers for future use
		if (savename)
			numbers. Save (savename);
	}

	// run as a worker if requested to
	else
	{
		std::cout << "Running as a worker..." << std::endl;
		int result = numbers. Work ();
		if (result == mwOk)
			std::cout << "Done." << std::endl;
		else
			std::cout << "Could not work - probably an error "
				"occurred." << std::endl;
	}

	return 0;	
} /* main */

