/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file cubisets.cpp
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: June 3, 2007.


#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/homology/chains.h"
#include "chomp/struct/bitfield.h"
#include "chomp/struct/integer.h"
#include "chomp/struct/hashsets.h"
#include "chomp/homology/gcomplex.h"
#include "chomp/homology/cubisets.h"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <cstdlib>
#include <cctype>

namespace chomp {
namespace homology {


// --------------------------------------------------
// ---------------------- BDDs ----------------------
// --------------------------------------------------

int MaxBddDim = MaxBddDimPossible;

bool UseMalandainCode = false;


// --------------------------------------------------
// ------------------- neighbors --------------------
// --------------------------------------------------

int getmaxneighbors (int dim)
{
	static int dimension = 0;
	static int maxn = 0;

	if (dim == dimension)
		return maxn;

	maxn = 1;
	for (int i = 0; i < dim; i ++)
		maxn *= 3;
	maxn --;

	dimension = dim;

	return maxn;
} /* getmaxneighbors */


// --------------------------------------------------
// ------------------ reduce cubes ------------------
// --------------------------------------------------

inline Cube zerocube (int dim)
{
	coordinate c [Cube::MaxDim];
	for (int i = 0; i < dim; ++ i)
		c [i] = 0;
	Cube q (c, dim);
	return q;
} /* zerocube */

int acyclic (int dim, BitField &b)
{
	// look for the answer in the tabulated data
	const char *table = tabulated [dim];
	if (table)
		return Tabulated::get (table,
			bits2int (b, getmaxneighbors (dim)));

	// look for the answer among the known combinations
	SetOfBitFields *known = knownbits [dim];
	int answer = known ? known -> check (b) : -1;

	// if the answer is known, return it
	if (answer >= 0)
		return answer;

	// create a model cube for building the neighborhood
	Cube qzero = zerocube (dim);

	// find out whether the set of neighbors is acyclic or not
/*	SetOfCubes neighbors;
	addneighbors (qzero, b, neighbors);
	SetOfCubes empty;
	cubreducequiet (neighbors, empty, empty, known);
	answer = (neighbors. size () == 1);
*/
// /*
	gcomplex<CubicalCell,integer> neighbors;
	addneighbors (qzero, b, neighbors, true);
	answer = (int) acyclic (neighbors);
//*/
	// add the answer to the list of known ones
	if (known)
		known -> add (b, answer);

	return answer;
} /* acyclic */

// --------------------------------------------------

int Tabulated::compute (int dim)
{
	// if the dimension is wrong, do nothing
	if ((dim <= 0) || (dim >= maxdim))
		return -1;

	// allocate memory for the table if necessary
	if (!tables [dim])
	{
		tables [dim] = new char [size [dim]];
		deallocate [dim] = true;
	}

	// remember the memory buffer and temporarily set no tabulating
	char *thetable = tables [dim];
	tables [dim] = 0;

	// clear the entire table
	for (int i = 0; i < size [dim]; ++ i)
		thetable [i] = 0;

	// prepare a bit field
	int maxneighbors = getmaxneighbors (dim);
	int nconfigs = 1 << maxneighbors;
	BitField b;
	b. allocate (maxneighbors);

	// prepare a temporary cube
	coordinate c [Cube::MaxDim];
	for (int i = 0; i < dim; ++ i)
		c [i] = 0;
	Cube qzero (c, dim);

	// compute all the configurations
	sout << "Computing " << dim << "-dimensional configurations... ";
	for (int i = 1; i < nconfigs; ++ i)
	{
		int2bits (i, maxneighbors, b);
	//	gcomplex<CubicalCell,integer> neighbors;
	//	addneighbors (qzero, b, neighbors, true);
	//	if (acyclic (neighbors))
		SetOfCubes neighbors;
		addneighbors (qzero, b, neighbors, true);
		if (acyclic (qzero, dim, neighbors, &b, maxneighbors))
			Tabulated::set (thetable, i);
		if (!(i % 153791))
			scon << std::setw (12) << i <<
				"\b\b\b\b\b\b\b\b\b\b\b\b";
	}
	sout << "Completed successfully.\n";

	// deallocate the bit field
	b. free ();

	// set the tabulating table to the just computed one
	tables [dim] = thetable;
	return 0;
} /* Tabulated::compute */



} // namespace homology
} // namespace chomp

// --------------------------------------------------


/// @}

