/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file mw.cpp
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2005 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on August 11, 2004. Last revision: January 13, 2006.

#include "chomp/multiwork/mw.h"
#include "chomp/multiwork/mw_none.h"
#include "chomp/multiwork/mw_unix.h"
#include "chomp/multiwork/mw_win32.h"
#include "chomp/multiwork/mw_wxwin.h"

// --------------------------------------------------
// ------------------- debugging --------------------
// --------------------------------------------------

//#include <iostream>
//using std::cerr;
//using std::endl;


namespace chomp {
namespace multiwork {

// --------------------------------------------------
// ------------- network communication --------------
// --------------------------------------------------
// The following functions are implemented
// in a system-dependent way in the "mw_*.cpp" files
// one of which (or all) must be linked with each program.

// Send the given buffer to the given socket.
// Return mwOk, mwError or mwLost.
int mwSendBytes (int fd, const unsigned char *buf, int len);

// Receive the given amount of data from the given socket.
// Return mwOk, mwError or mwLost.
int mwRecvBytes (int fd, unsigned char *buf, int len);

// Connect to the given computer at the given port.
// Return the socket number (non-negative) or mwError.
int mwConnect (const char *name, int port);

// Begin listening at the given port. Allow for the given queue length.
// Return the socket number (non-negative) or mwError.
int mwListen (int port, int queuesize);

// Wait for and accept a connection at the given socket.
// Return the new socket number (non-negative), mwTimeOut or mwError.
int mwAccept (int fd, mwData &computer, int timeout = -1);

// Determine IOflags for each of the workers + listensocket (the last flag).
// The IOflags of the interest are initially set.
// Return mwOk, mwTimeOut or mwError.
int mwSelect (const int *workers, int nworkers, int listensocket,
	int *ioflags, int timeout);

// Disconnect the given socket. If negative number, this is ignored.
void mwDisconnect (int fd);


// --------------------------------------------------
// -------------------- messages --------------------
// --------------------------------------------------

static int mwSendMessage (int fd, unsigned int code, const mwData &x)
// Send a message to the given socket.
// Return mwOk, mwError or mwLost.
{
	mwData sending;
	sending << code << (int) x;
	sending << x;
	return mwSendBytes (fd, sending, sending);
} /* mwSendMessage */

static int mwRecvMessage (int fd, unsigned int &code, mwData &x)
// Receive a message from the given socket.
// Return mwOk, mwError or mwLost.
{
	// read the code and length of the message
	unsigned char buf0 [8];
	int result = mwRecvBytes (fd, buf0, 8);
	if (result < 0)
		return result;

	// extract the length of the message
	int len = (int) (buf0 [4]) << 24;
	len |= (int) (buf0 [5]) << 16;
	len |= (int) (buf0 [6]) << 8;
	len |= buf0 [7];
	if (len < 0)
		return mwError;

	// extract the code of the message
	code = (int) (buf0 [0]) << 24;
	code |= (int) (buf0 [1]) << 16;
	code |= (int) (buf0 [2]) << 8;
	code |= buf0 [3];

	// if the message length is zero, no more reading is necessary
	if (!len)
	{
		x. Reset ();
		return mwOk;
	}

	// prepare a memory buffer for the message
	unsigned char *buf = new unsigned char [len];
	if (!buf)
		return mwError;

	// read the message
	result = mwRecvBytes (fd, buf, len);
	if (result < 0)
	{
		delete [] buf;
		return result;
	}

	// transform the message to mw data
	x. Take (buf, len);
	return mwOk;
} /* mwRecvMessage */


// --------------------------------------------------
// --------------------- tables ---------------------
// --------------------------------------------------

inline void mwTableDel (int *tab, int len, int pos)
// Delete the entry at position 'pos' from the table of 'len' positions.
// Shift all the positions after 'pos' to the back.
{
	for (int i = pos + 1; i < len; i ++)
		tab [i - 1] = tab [i];
	return;
} /* mwTableDel */

inline void mwTableDel (mwData *tab, int len, int pos)
// Delete the entry at position 'pos' from the table of 'len' positions.
// Shift all the positions after 'pos' to the back.
{
	for (int i = pos + 1; i < len; i ++)
		tab [i - 1]. Take (tab [i]);
	return;
} /* mwTableDel */


// --------------------------------------------------
// --------------------- mwTask ---------------------
// --------------------------------------------------

int mwTask::QuitWorkers ()
{
	// prepare a char buffer long enough for any name
	char *name = new char [(int) computers];
	if (!name)
		return mwError;

	// rewind the buffer with computer names to the beginning
	computers. Rewind ();

	// write to the log file what you are doing
	Log ("Turning workers off...");
	
	// try connecting to each worker and ask them to exit
	int n = 0;
	while (true)
	{
		// prepare empty data (if cannot read from the list)
		int port = 0;
		*name = '\0';

		// retrieve the computer name and port from the list
		computers >> name >> port;
	
		// if no name or no port read, quit the loop
		if (!*name || !port)
			break;

		// try connecting to the computer
		int fd = mwConnect (name, port);

		// if unsuccessful, make a note and take another one
		if (fd < 0)
		{
			if (logFile)
				Log (*logFile << "Worker " << name << ":" <<
					port << " could not be contacted.");
			continue;
		}
		
		// prepare the control code to send to the worker
		unsigned int code = (ctrlnum << mwShiftMsg) | mwByeMsg |
			mwDontKeepMsg;

		// send the 'Bye!' message to the worker and disconnect it
		mwData empty;
		int result = mwSendMessage (fd, code, empty);
		mwDisconnect (fd);

		// add an appropriate message to the log file
		if (logFile && (result == mwOk))
			Log (*logFile << "Worker " << name << ":" <<
				port << " exited successfully.");
		else if (logFile && (result != mwOk))
			Log (*logFile << "Error while sending the 'Bye!' "
				"message to " << name << ":" <<
				port << ".");

		// increase the workers' counter
		n ++;
	}

	// write to the log file how many workers were turned off
	if (logFile)
		Log (*logFile << n << " workers have been shut down.");
	
	// get rid of the temporary char buffer
	delete [] name;

	return mwOk;
} /* mwTask::QuitWorkers */


// --------------------------------------------------
// -------------------- mwWorker --------------------
// --------------------------------------------------

int mwWorker::Work ()
{
	Log ("Running as a WORKER.");

	// indicate what kind of network connection is in use
	#if !mwNETWORK
	Log ("There is no network in use, so exiting right now.");
	#elif mwWXWIN
	Log ("Using the sockets interface provided by wxWindows.");
	#else
	Log ("Using the standard sockets for network connections.");
	#endif

	// remove unnecessary bits from the control number
	ctrlnum <<= mwShiftMsg;
	ctrlnum >>= mwShiftMsg;
	negctrlnum <<= mwShiftMsg;
	negctrlnum >>= mwShiftMsg;

	// connect to the first comp. on the list that is running
	int fd = -1;
	if ((int) computers)
	{
		// prepare a char buffer long enough for any name
		char *name = new char [(int) computers];
		if (!name)
			return mwError;

		// rewind the buffer with computer names to the beginning
		computers. Rewind ();

		// try connecting until succeeded
		while (fd < 0)
		{
			// prepare empty data (if cannot read from the list)
			int port = 0;
			*name = '\0';

			// retrieve the computer name and port from the list
			computers >> name >> port;
	
			// if no name or no port read, quit the loop
			if (!*name || !port)
				break;

			// try connecting to the computer
			fd = mwConnect (name, port);

			// add an appropriate message to the log file
			if (logFile)
				Log (*logFile << "Connection to " <<
					name << ":" << port << ((fd < 0) ?
					" refused." : " established."));
		}

		// get rid of the temporary char buffer
		delete [] name;
	}

	// be ready to send the port number at which the worker is listening
	bool sendPort = (portnum > 0) && (fd >= 0);

	while (1)
	{
		// prepare a socket number at which to listen
		int fdlisten = -1;

		// if not connected and not listening, begin listening
		if ((fd < 0) && (portnum > 0) && (fdlisten < 0))
		{
			// add a suitable message to the log
			if (logFile)
				Log (*logFile << "Waiting for a coordinator "
					"at port " << portnum << ".");

			// listen at the given port (use a very short queue)
			fdlisten = mwListen (portnum, 1);

			// if the listening failed, return with an error
			if (fdlisten < 0)
			{
				Log ("Error: This port is probably in use.");
				return mwError;
			}
		}

		// if not connected and listening, try connecting
		if ((fd < 0) && (fdlisten >= 0))
		{
			// accept a connection and stop listening
			mwData computer;
			fd = mwAccept (fdlisten, computer, timeout);
			mwDisconnect (fdlisten);
			fdlisten = -1;
	
			// if too much time elapsed, quit the job
			if (fd == mwTimeOut)
			{
				Log ("Time-out. Quitting the job.");
				return mwOk;
			}

			// if an error occurred, quit the job
			else if (fd < 0)
			{
				Log ("Error while connecting "
					"to a coordinator.");
				return mwError;
			}

			// report the fact of connection to the log file
			if (logFile)
				Log (*logFile << "Connected to a "
					"coordinator at " <<
					(const char *) computer << ".");
		}

		// receive messages and work
		while (1)
		{
			// receive a message
			mwData msg;
			unsigned int code = 0;
			int result = mwRecvMessage (fd, code, msg);

			// disconnect in case of an I/O error
			if (result == mwError)
			{
				Log ("Error while receiving data.");
				mwDisconnect (fd);
				fd = -1;
				if (keepWorker)
					break;
				else
					return mwError;
			}
	
			// disconnect if the connection was lost
			if (result == mwLost)
			{
				Log ("The connection closed by peer.");
				mwDisconnect (fd);
				fd = -1;
				if (keepWorker)
					break;
				else
					return mwOk;
			}
	
			// disconnect if wrong code received
			if ((code >> mwShiftMsg) != ctrlnum)
			{
				Log ("Wrong control code. Disconnecting.");
				mwDisconnect (fd);
				fd = -1;
				if (keepWorker)
					break;
				else
					return mwError;
			}

			// update the 'keepWorker' flag if requested to
			if (code & mwKeepMsg)
				keepWorker = true;
			if (code & mwDontKeepMsg)
				keepWorker = false;

			// disconnect if requested to
			if (code & mwByeMsg)
			{
				Log ("Disconnecting on request.");
				mwDisconnect (fd);
				fd = -1;
				if (keepWorker)
					break;
				else
					return mwOk;
			}

			// process an initialization message
			if (code & mwInitMsg)
			{
				if (Initialize (msg) == mwError)
				{
					Log ("Bad initialization data.");
					mwDisconnect (fd);
					return mwError;
				}
				if (logLevel >= mwDetailed)
					Log ("* Initialization data "
						"processed.");
			}

			// process a standard message
			if (code & mwStdMsg)
			{
				// prepare a basis for the returned code
				unsigned int retcode = negctrlnum <<
					mwShiftMsg;
		
				// process the message data
				int result = Process (msg);
		
				// if rejected, reset the data and set flag
				if (result == mwReject)
				{
					msg. Reset ();
					retcode |= mwRejectedMsg;
				}

				// if an error occurred, disconnect and quit
				else if (result == mwError)
				{
					Log ("Data processing failed.");
					mwDisconnect (fd);
					fd = -1;
					return mwError;
				}

				// write what was done to the report file
				if (logLevel >= mwDetailed)
					Log ((retcode & mwRejectedMsg) ?
						"* Data rejected." :
						"* Data processed.");

				// send port number if relevant
				if (sendPort)
				{
					mwData d;
					d << portnum;
					unsigned code = mwPortMsg |
						(negctrlnum << mwShiftMsg);
					mwSendMessage (fd, code, d);
					sendPort = false;
					if (logLevel >= mwDetailed)
						Log ("* Port number sent.");
				}

				// send the result of the processed data
				result = mwSendMessage (fd, retcode, msg);

				// exit the loop if the connection was lost
				if (result == mwLost)
				{
					Log ("Connection lost "
						"while sending data.");
					mwDisconnect (fd);
					fd = -1;
					if (keepWorker)
						break;
					else
						return mwError;
				}
		
				// exit the loop if an error occurred
				if (result != mwOk)
				//if (result == mwError)
				{
					Log ("Error while sending data.");
					mwDisconnect (fd);
					fd = -1;
					if (keepWorker)
						break;
					else
						return mwError;
				}
			}
		}

		// add a line to the log file between working sessions
		if ((fd < 0) && (portnum > 0))
			Log ("============================");
	}
} /* mwWorker::Work */


// --------------------------------------------------
// ----------------- mwCoordinator ------------------
// --------------------------------------------------

int mwCoordinator::Run (int options)
{
	// prepare the flags that are going to be returned
	int flags = mwNone;

	// if there are any rejected portions of data, set this flag
	if (nRejected)
		flags |= mwRejected;

	// if some results are ready, set this flag
	if (nDone)
		flags |= mwArrived;

	// if no work, ask for more and say that work is completed
	if (!nWork)
	{
		if (!(options & mwNoMoreData))
			flags |= mwHungry;
		flags |= mwCompleted;
	}

	if (badInitialization)
		flags |= mwFailed;

	// if there is data to read or no work to do, exit now
	if (nRejected || nDone || !nWork || badInitialization)
		return flags;

	if (singleWork && localWorker)
	{
		// log the basic information to the log file
		if (firstrun)
		{
			Log ("Running in the single-work mode.");
			firstrun = false;
		}

		// process all the initialization data pending
		while (singleinitsent < nInit)
		{
		//	xInit [singleinitsent]. Rewind ();
			if (localWorker -> Initialize (xInit
				[singleinitsent ++]) != mwOk)
			{
				badInitialization = true;
				return (flags | mwFailed);
			}
		}

		// process the data
		//while (nWork && !nDone && !nRejected)
		while (nWork && (nDone < mwMAXDATA) &&
			(nRejected < mwMAXDATA))
		{
			// copy the data to the place for the result
			xDone [nDone] = xWork [-- nWork];

			// process the data and replace it with the result
			xDone [nDone]. Rewind ();
			int result = localWorker -> Process (xDone [nDone]);
			if (result == mwReject)
			{
				// if the data was rejected, move it there
				xRejected [nRejected ++].
					Take (xWork [nWork]);
				flags |= mwRejected;
			}
			else if (result == mwError)
			{
				Log ("Data processing failed.");
				flags |= mwFailed;
			}
			else
			{
				// if the data was processed, it is there
				nDone ++;
				flags |= mwArrived;
			}
		}

		// return the flags
		return flags;

	} // end of singleWork

	if (firstrun)
	{
		Log ("Running as a COORDINATOR.");

		// indicate what kind of network connection is in use
		#if !mwNETWORK
		Log ("There is no network in use.");
		#elif mwWXWIN
		Log ("Using the sockets interface provided by wxWindows.");
		#else
		Log ("Using the standard sockets for network connections.");
		#endif
	}

	// connect to workers who are waiting if this is the first run
	if (firstrun && (int) computers)
	{
		char *name = new char [(int) computers];
		computers. Rewind ();
		while (nworkers <= mwMAXWORK)
		{
			// retrieve the computer name and port number
			if (!name)
				break;
			int port = 0;
			*name = '\0';
			computers >> name >> port;
			if (!*name || !port)
				break;

			// connect to the computer
			int fd = mwConnect (name, port);

			// if the connection succeeded, add this worker
			if (fd >= 0)
			{
				if (logFile)
					Log (*logFile << "Connected to " <<
						name << ":" << port);
				sentdata [nworkers] = -1;
				sentinit [nworkers] = 0;
				names [nworkers]. Reset ();
				names [nworkers] << name;
				ports [nworkers] = port;
				workers [nworkers ++] = fd;
				nworking ++;
			}

			// if unable to connect, make only a note in the log
			else if (logFile)
				Log (*logFile << "Connection attempt to " <<
					name << ":" << port << " failed.");
		}
		delete [] name;
	}

	// begin listening at the given port for more incoming workers
	if (firstrun && (portnum > 0))
	{
		listensocket = mwListen (portnum, 5);
		if (logFile)
			if (listensocket >= 0)
				Log (*logFile << "Waiting for workers "
					"at port " << portnum << ".");
			else
				Log (*logFile << "Listening attempt at "
					"port " << portnum << " failed.");
	}

	// if not listening and there are no workers...
	if ((listensocket < 0) && !nworking)
	{
		// switch to the single-work mode if allowed to
		if (localWorker)
		{
			singleWork = true;
			if (firstrun)
				Log ("There are no workers: Switching "
					"to the single-work mode.");
			else
				Log ("All workers disconnected: Switching "
					"to the single-work mode.");
		}
		// if data cannot be processed locally, report a failure
		else
		{
			Log ("Failure: The work cannot be done.");
			flags |= mwFailed;
		}
		return flags;
	}

	// make a note that this is no longer going to be the first run
	firstrun = false;

	// process data to get at least one piece completed
	//while (nWork && (nDone < mwMAXDATA) && (nRejected < mwMAXDATA))
	while (!nDone && !nRejected)
	// note: in the loop, nWork + nDone + nRejected = const > 0
	{
		// TODO: If there is no more data (options & mwNoMoreData)
		// and some workers are idle, and some others are working
		// for a long time, send their data also to a few idle
		// workers... This requires some A.I., of course. ;)

		int i;
		// determine whether there is still something to send
		int sentcount = 0;
		for (i = 0; i < nworkers; ++ i)
			if (sentdata [i] >= 0)
				++ sentcount;
		bool worktodo = (nWork > sentcount) ||
			!(options & mwNoMoreData);

		// prepare the listen-socket flag if we are interested in it
		if ((listensocket >= 0) && worktodo)
			ioflags [nworkers] = mwCanRead;
		else
			ioflags [nworkers] = mwNone;

		// prepare these workers' flags which we are interested in
		for (i = 0; i < nworkers; ++ i)
		{
			if (sentdata [i] != -1)
				ioflags [i] = mwCanRead;
			else if (worktodo && !(workers [i] < 0))
				ioflags [i] = mwCanWrite;
			else
				ioflags [i] = mwNone;
		}

		// determine whether it is necessary to wait or not
		int t = timeout;
		if (localWorker && !nworking && !alldatasent)
			t = 0;

		// report the select's parameters to the log file
		if (logFile && (logLevel >= mwDebug))
		{
			*logFile << ">>> Select, t = " << t << ", flags =";
			for (i = 0; i <= nworkers; ++ i)
				*logFile << " " << ioflags [i];
			if (t)
				Log (".");
		}
	
		// wait until data can be received from or sent to any socket
		int result = mwSelect (workers, nworkers, listensocket,
			ioflags, t);

		// report the returned parametrs to the log file
		if (logFile && (logLevel >= mwDebug))
		{
			if (t)
				*logFile << ">>> Select returned flags =";
			else
				*logFile << ", returned =";
			for (i = 0; i <= nworkers; ++ i)
				*logFile << " " << ioflags [i];
			Log (".");
		}

		// in case of select's failure, close all the connections
		if (result == mwError)
		{
			// report the failure to the log file
			Log ("The 'select' function failed badly.");

			// disconnect all the workers
			for (int i = 0; i < nworkers; i ++)
				mwDisconnect (workers [i]);

			// shut down the listening socket
			if (listensocket >= 0)
				mwDisconnect (listensocket);
			nworkers = 0;

			// switch to the single-work mode if allowed to
			if (localWorker)
			{
				Log ("Switching to the single-work mode.");
				singleWork = true;
			}
			else
			{
				Log ("Reporting a failure condition.");
				flags |= mwFailed;
			}
				
			break;
		}

		// exit the loop if a time-out occurred
		if (!localWorker && (result == mwTimeOut))
		{
			if (logLevel >= mwDebug)
				Log ("Time-out occurred at 'select'.");
			break;
		}

		// receive data from all the workers
		for (i = 0; (i < nworkers) && (nDone < mwMAXDATA) &&
			(nRejected < mwMAXDATA); ++ i)
		{
			// if this worker has no data, skip it
			if (!(ioflags [i] & mwCanRead))
				continue;

			// receive the entire data chunk from the worker
			unsigned int code = 0;
			int result = mwRecvMessage (workers [i], code,
				xDone [nDone]);

			// reject the worker in case of error or wrong code
			if ((result < 0) || ((code >> mwShiftMsg) !=
				((negctrlnum << mwShiftMsg) >> mwShiftMsg)))
			{
				// log the details of what happened
				if (logFile)
				{
					*logFile << "Worker " << i;
					if (*(const char *) (names [i]))
						*logFile << " (" <<
							(const char *)
							(names [i]) << ')';
					*logFile << " disconnected: ";
					if (result == mwLost)
						Log ("Connection lost.");
					else if (result < 0)
						Log ("An error occurred.");
					else
						Log ("Wrong code received.");
				}

				// disconnect the worker
				mwDisconnect (workers [i]);

				// reset the worker's data
				workers [i] = -1;
				sentdata [i] = -1;
				-- nworking;
				continue;
			}

			// if transmitting the port number only, take it
			if (code & mwPortMsg)
			{
				// retrieve the port number
				xDone [nDone] >> ports [i];

				// report this fact to the log file
				if (logFile && (logLevel >= mwDetailed))
					Log (*logFile << "Port number " <<
						ports [i] << " received "
						"from worker " << i << '.');
				continue;
			}
	
			// determine the worker's data piece number
			int nr = sentdata [i];
	
			// this worker is ready for another piece of data
			sentdata [i] = -1;

			// if the data is to be ignored, skip the rest
			if (nr == -2)
				continue;

			// if the data was rejected, move it to 'xRejected'
			if (code & mwRejectedMsg)
			{
				// report what happened
				if (logFile && (logLevel >= mwDetailed))
					Log (*logFile << "Data was rejected "
						"by worker " << i << '.');

				// move the data to the rejected table
				if (nRejected < mwMAXDATA)
					xRejected [nRejected ++].
						Take (xWork [nr]);

				// set an appropriate flag
				flags |= mwRejected;
			}

			// if the data was Ok, the result is in 'xDone'
			else
			{
				// report what happened
				if (logFile && (logLevel >= mwDetailed))
					Log (*logFile << "Processed data "
						"arrived from worker " <<
						i << '.');

				// accept this piece of data
				nDone ++;

				// set an appropriate flag
				flags |= mwArrived;
			}

			// remove the processed data from the 'xWork' table
			mwTableDel (xWork, nWork --, nr);

			// set the "completed" flag if relevant
			if (!nWork)
				flags |= mwCompleted;

			// update the indices of sent data
			for (int j = 0; j < nworkers; j ++)
			{
				if (sentdata [j] > nr)
					sentdata [j] --;
				else if (sentdata [j] == nr)
					sentdata [j] = -2;
			}
		}

		// create a table indicating how many workers have each data
		int j;
		int sent [mwMAXDATA];
		for (j = 0; j < mwMAXDATA; ++ j)
			sent [j] = 0;
		for (j = 0; j < nworkers; ++ j)
			if (sentdata [j] >= 0)
				sent [sentdata [j]] ++;

		// reset the number of the first data available for sending
		int first = 0;

		// send data if possible
		for (i = 0; i < nworkers; ++ i)
		{
			// if this worker is not ready, skip it
			if ((sentdata [i] != -1) ||
				!(ioflags [i] & mwCanWrite))
				continue;

			// send initial data if any is pending
			int result = mwOk;
			if (sentinit [i] < nInit)
			{
				unsigned int code = (ctrlnum << mwShiftMsg) |
					mwInitMsg;
				if (keepWorkers)
					code |= mwKeepMsg;
				else
					code |= mwDontKeepMsg;
				result = mwSendMessage (workers [i], code,
					xInit [sentinit [i] ++]);
				if (result == mwOk)
				{
					if (logFile &&
						(logLevel >= mwDetailed))
						Log (*logFile << "Initial "
							"data " <<
							(sentinit [i] - 1) <<
							" sent to worker " <<
							i << '.');
					continue;
				}
			}

			// determine the first data available for sending
			while ((first < nWork) && (sent [first]))
				++ first;

			// if there is no data for sending, say "hungry"
			if ((first >= nWork) && (nWork < mwMAXWORK) &&
				(result == mwOk))
				flags |= mwHungry;

			// send a data chunk for processing if available
			if ((first < nWork) && (result == mwOk))
			{
				unsigned int code = (ctrlnum << mwShiftMsg) |
					mwStdMsg;
				if (keepWorkers)
					code |= mwKeepMsg;
				else
					code |= mwDontKeepMsg;
				result = mwSendMessage (workers [i], code,
					xWork [first]);
				if (result == mwOk)
				{
					if (logFile &&
						(logLevel >= mwDetailed))
						Log (*logFile << "Data sent "
							"to worker " << i <<
							'.');
					sentdata [i] = first;
					sent [first] ++;
					continue;
				}
			}
	
			// if an error occurred, reject the worker
			if (result != mwOk)
			{
				if (logFile)
					Log (*logFile << "Worker " << i <<
						" disconnected: " <<
						((result == mwLost) ?
						"Connection lost." :
						"An error occurred."));

				// disconnect the worker
				mwDisconnect (workers [i]);

				// reset the worker's data
				workers [i] = -1;
				sentdata [i] = -1;
				-- nworking;
			}
		}

		// if the task has to be run locally, do it now;
		// note: I use 'while' instead of 'if' in order to do 'break'
		while (localWorker && !nworking && !alldatasent &&
			(result == mwTimeOut))
		{
			// determine the first data available for sending
			while ((first < nWork) && (sent [first]))
				++ first;

			// process a piece of the initialization data if any
			if (singleinitsent < nInit)
			{
			//	xInit [singleinitsent]. Rewind ();
				if (localWorker -> Initialize (xInit
					[singleinitsent ++]) != mwOk)
				{
					badInitialization = true;
					flags |= mwFailed;
				}
				break;
			}

			// if no work is waiting to be done, ask for more
			if (first >= nWork)
			{
				alldatasent = true;
				if (!(options & mwNoMoreData))
					flags |= mwHungry;
				break;
			}

			// if there is no room for results, don't work
			if (nDone >= mwMAXDATA)
				break;

			// make a note in the log file of what is going on
			if (logLevel >= mwDetailed)
				Log ("Processing a piece of data locally.");
	
			// process one piece of data
			mwData current (xWork [first]);
			current. Rewind ();
			int result = localWorker -> Process (current);
			if (result == mwReject)
			{
				if (nRejected < mwMAXDATA)
					xRejected [nRejected ++].
						Take (xWork [first]);
				flags |= mwRejected;
			}
			else if (result == mwError)
			{
				Log ("Data processing failed.");
				flags |= mwFailed;
			}
			else
			{
				xDone [nDone ++]. Take (current);
				flags |= mwArrived;
			}

			// remove the processed data from the 'xWork' table
			int j;
			-- nWork;
			for (j = first; j < nWork; ++ j)
				xWork [j]. Take (xWork [j + 1]);

			// set the "completed" flag if relevant
			if (!nWork)
				flags |= mwCompleted;

			// update the indices of sent data
			for (j = 0; j < nworkers; j ++)
				if (sentdata [j] > first)
					sentdata [j] --;

			// exit the 'while' loop (sorry for this hack!)
			break;
		}

		// accept new connections if any
		if ((ioflags [nworkers] & mwCanRead) &&
			(nworkers < mwMAXWORK))
		{
			// accept the connection
			names [nworkers]. Reset ();
			workers [nworkers] = mwAccept (listensocket,
				names [nworkers]);

			// if accepted successfully...
			if (workers [nworkers] >= 0)
			{
				// log the worker's acceptance
				if (logFile)
				{
					*logFile << "Worker " << nworkers;
					if ((const char *)(names [nworkers]))
						*logFile << " from " <<
							(const char *) (names
							[nworkers]);
					Log (*logFile << " accepted.");
				}

				// initialization data has to be sent
				if (nInit)
					alldatasent = false;

				// add the worker and reset its data
				sentinit [nworkers] = 0;
				sentdata [nworkers] = -1;
				ports [nworkers ++] = -1;
				++ nworking;
			}
			// report the error to the log file
			else
				Log ("Unsuccessful connection of a worker.");
		}

		// if new data can be sent, exit the loop
		if (flags & mwHungry)
			break;
	}

	return flags;
} /* mwCoordinator::Run */

// --------------------------------------------------

void mwCoordinator::Disconnect ()
{
	// release all the workers if they are still connected
	for (int i = 0; i < nworkers; ++ i)
	{
		// if this worker is already disconnected, take the next one
		if (workers [i] < 0)
			continue;

		// prepare the control code to send to the worker
		unsigned int code = (ctrlnum << mwShiftMsg) | mwByeMsg;
		if (keepWorkers)
			code |= mwKeepMsg;
		else
			code |= mwDontKeepMsg;

		// send the 'Bye!' message to the worker
		mwData empty;
		mwSendMessage (workers [i], code, empty);

		// disconnect the worker
		mwDisconnect (workers [i]);
		workers [i] = -1;

		// make a note of what happened in the log file
		if (logFile)
		{
			*logFile << "Worker " << i;
			if (*(const char *) (names [i]))
			{
				*logFile << " (" <<
					(const char *) (names [i]);
				if (ports [i] > 0)
					*logFile << ':' << ports [i];
				*logFile << ')';
			}
			*logFile << " disconnected and " <<
				(keepWorkers ? "is waiting." : "exited.");
			Log ();
		}
	}

	return;
} /* mwCoordinator::Disconnect */


} // namespace multiwork
} // namespace chomp

/// @}

